# -*- coding: utf-8 -*-
"""
==================
MTH5
==================

MTH5 deals with reading and writing an MTH5 file, which are HDF5 files
developed for magnetotelluric (MT) data.  The code is based on h5py and
therefor numpy.  This is the simplest and we are not really dealing with
large tables of data to warrant using pytables.

Created on Sun Dec  9 20:50:41 2018

:copyright: Jared Peacock (jpeacock@usgs.gov)
    
:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================
from pathlib import Path
from platform import platform

import h5py

from mth5.utils.exceptions import MTH5Error
from mth5 import __version__ as mth5_version
from mth5 import groups as groups
from mth5.tables import ChannelSummaryTable, TFSummaryTable
from mth5 import helpers
from mth5.utils.mth5_logger import setup_logger
from mth5 import CHANNEL_DTYPE, TF_DTYPE

from mt_metadata.utils.mttime import get_now_utc
from mt_metadata.timeseries import Experiment
from mt_metadata.transfer_functions.core import TF

# =============================================================================
# Acceptable parameters
# =============================================================================
acceptable_file_types = ["mth5", "MTH5", "h5", "H5"]
acceptable_file_versions = ["0.1.0", "0.2.0"]
acceptable_data_levels = [0, 1, 2, 3]

# =============================================================================
# MT HDF5 file
# =============================================================================


class MTH5:
    """
    MTH5 is the main container for the HDF5 file format developed for MT data

    It uses the metadata standards developled by the
    `IRIS PASSCAL software group
    <https://www.iris.edu/hq/about_iris/governance/mt_soft>`_
    and defined in the
    `metadata documentation
    <https://github.com/kujaku11/MTarchive/blob/tables/docs/mt_metadata_guide.pdf>`_.

    MTH5 is built with h5py and therefore numpy.  The structure follows the
    different levels of MT data collection:

    For version 0.1.0:

        - Survey

           - Reports
           - Standards
           - Filters
           - Stations

               - Run

                   - Channel

    For version 0.2.0:

        - Experiment

            - Reports
            - Standards
            - Surveys

               - Reports
               - Standards
               - Filters
               - Stations

                   - Run

                       -Channel


    All timeseries data are stored as individual channels with the appropriate
    metadata defined for the given channel, i.e. electric, magnetic, auxiliary.

    Each level is represented as a mth5 group class object which has methods
    to add, remove, and get a group from the level below.  Each group has a
    metadata attribute that is the approprate metadata class object.  For
    instance the SurveyGroup has an attribute metadata that is a
    :class:`mth5.metadata.Survey` object.  Metadata is stored in the HDF5 group
    attributes as (key, value) pairs.

    All groups are represented by their structure tree and can be shown
    at any time from the command line.

    Each level has a summary array of the contents of the levels below to
    hopefully make searching easier.

    :param filename: name of the to be or existing file
    :type filename: string or :class:`pathlib.Path`
    :param compression: compression type.  Supported lossless compressions are

        * 'lzf' - Available with every installation of h5py
                 (C source code also available). Low to
                 moderate compression, very fast. No options.
        * 'gzip' - Available with every installation of HDF5,
                  so it’s best where portability is required.
                  Good compression, moderate speed.
                  compression_opts sets the compression level
                  and may be an integer from 0 to 9,
                  default is 3.
        * 'szip' - Patent-encumbered filter used in the NASA
                   community. Not available with all
                   installations of HDF5 due to legal reasons.
                   Consult the HDF5 docs for filter options.

    :param compression_opts: compression options, see above
    :type compression_opts: string or int depending on compression type
    :param shuffle: Block-oriented compressors like GZIP or LZF work better
                    when presented with runs of similar values. Enabling the
                    shuffle filter rearranges the bytes in the chunk and may
                    improve compression ratio. No significant speed penalty,
                    lossless.
    :type shuffle: boolean
    :param fletcher32: Adds a checksum to each chunk to detect data corruption.
                       Attempts to read corrupted chunks will fail with an
                       error. No significant speed penalty. Obviously
                       shouldn’t be used with lossy compression filters.
    :type fletcher32: boolean
    :param data_level: level the data are stored following levels defined by
       `NASA ESDS <https://earthdata.nasa.gov/collaborate/open-data-services-and-software/data-information-policy/data-levels>`_

         * 0 - Raw data
         * 1 - Raw data with response information and full metadata
         * 2 - Derived product, raw data has been manipulated

    :type data_level: integer, defaults to 1
    :param file_version: Version of the file [ '0.1.0' | '0.2.0' ], defaults to "0.2.0"
    :type file_version: string, optional

    :Usage:

    * Open a new file and show initialized file

    >>> from mth5 import mth5
    >>> mth5_obj = mth5.MTH5(file_version='0.1.0')
    >>> # Have a look at the dataset options
    >>> mth5.dataset_options
    {'compression': 'gzip',
     'compression_opts': 3,
     'shuffle': True,
     'fletcher32': True}
    >>> mth5_obj.open_mth5(r"/home/mtdata/mt01.mth5", 'w')
    >>> mth5_obj
    /:
    ====================
        |- Group: Survey
        ----------------
            |- Group: Filters
            -----------------
                --> Dataset: summary
                ......................
            |- Group: Reports
            -----------------
                --> Dataset: summary
                ......................
            |- Group: Standards
            -------------------
                --> Dataset: summary
                ......................
            |- Group: Stations
            ------------------
                --> Dataset: summary
                ......................


    * Add metadata for survey from a dictionary

    >>> survey_dict = {'survey':{'acquired_by': 'me', 'archive_id': 'MTCND'}}
    >>> survey = mth5_obj.survey_group
    >>> survey.metadata.from_dict(survey_dict)
    >>> survey.metadata
    {
    "survey": {
        "acquired_by.author": "me",
        "acquired_by.comments": null,
        "archive_id": "MTCND"
        ...}
    }

    * Add a station from the convenience function

    >>> station = mth5_obj.add_station('MT001')
    >>> mth5_obj
    /:
    ====================
        |- Group: Survey
        ----------------
            |- Group: Filters
            -----------------
                --> Dataset: summary
                ......................
            |- Group: Reports
            -----------------
                --> Dataset: summary
                ......................
            |- Group: Standards
            -------------------
                --> Dataset: summary
                ......................
            |- Group: Stations
            ------------------
                |- Group: MT001
                ---------------
                    --> Dataset: summary
                    ......................
                --> Dataset: summary
                ......................
    >>> station
    /Survey/Stations/MT001:
    ====================
        --> Dataset: summary
        ......................

    >>> data.schedule_01.ex[0:10] = np.nan
    >>> data.calibration_hx[...] = np.logspace(-4, 4, 20)

    .. note:: if replacing an entire array with a new one you need to use [...]
              otherwise the data will not be updated.

    .. warning:: You can only replace entire arrays with arrays of the same
                 size.  Otherwise you need to delete the existing data and
                 make a new dataset.

    .. seealso:: https://www.hdfgroup.org/ and https://www.h5py.org/

    """

    def __init__(
        self,
        filename=None,
        compression="gzip",
        compression_opts=4,
        shuffle=True,
        fletcher32=True,
        data_level=1,
        file_version="0.2.0",
    ):

        self.logger = setup_logger(f"{__name__}.{self.__class__.__name__}")

        # make these private so the user cant accidentally change anything.
        self.__hdf5_obj = None
        (
            self.__compression,
            self.__compression_opts,
        ) = helpers.validate_compression(compression, compression_opts)
        self.__shuffle = shuffle
        self.__fletcher32 = fletcher32

        self.data_level = data_level
        self.filename = filename
        self.file_version = file_version
        self.file_type = "mth5"

        self._set_default_groups()

    def __str__(self):
        if self.h5_is_read():
            return helpers.get_tree(self.__hdf5_obj)
        return "HDF5 file is closed and cannot be accessed."

    def __repr__(self):
        return self.__str__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_mth5()
        return False

    @property
    def dataset_options(self):
        """summary of dataset options"""
        return {
            "compression": self.__compression,
            "compression_opts": self.__compression_opts,
            "shuffle": self.__shuffle,
            "fletcher32": self.__fletcher32,
        }

    @property
    def file_attributes(self):
        return {
            "file.type": "MTH5",
            "file.version": self.file_version,
            "file.access.platform": platform(),
            "file.access.time": get_now_utc(),
            "mth5.software.version": mth5_version,
            "mth5.software.name": "mth5",
            "data_level": self.data_level,
        }

    @property
    def filename(self):
        """file name of the hdf5 file"""
        if self.h5_is_read():
            return Path(self.__hdf5_obj.filename)
        msg = (
            "MTH5 file is not open or has not been created yet. "
            + "Returning default name"
        )
        self.logger.warning(msg)
        return self.__filename

    @filename.setter
    def filename(self, value):
        """make sure file has the proper extension"""
        self.__filename = None
        if value is not None:
            if not isinstance(value, Path):
                value = Path(value)
            if value.suffix not in acceptable_file_types:
                msg = (
                    f"file extension {value.suffix} is not correct. "
                    "Changing to default .h5"
                )
                self.logger.info(msg)
                self.__filename = value.with_suffix(".h5")
            else:
                self.__filename = value

    @property
    def file_type(self):
        """File Type should be MTH5"""

        if self.h5_is_read():
            # need the try statement for when a file is initialize it does
            # not have the attributes yet.
            try:
                return self.__hdf5_obj.attrs["file.type"]
            except KeyError:
                return self.__file_type
        return self.__file_type

    @file_type.setter
    def file_type(self, value):
        """set file type while validating input"""
        if not isinstance(value, str):
            msg = f"Input file type must be a string not {type(value)}"
            self.logger.error(msg)
            raise ValueError(msg)
        if value not in acceptable_file_types:
            msg = (
                f"Input file.type is not valid, must be {acceptable_file_types}"
            )
            self.logger.error(msg)
            raise ValueError(msg)
        self.__file_type = value

        if self.h5_is_read():
            self.__hdf5_obj.attrs["file.type"] = value

    @property
    def file_version(self):
        """mth5 file version"""
        if self.h5_is_read():
            # need the try statement for when a file is initialize it does
            # not have the attributes yet.
            try:
                return self.__hdf5_obj.attrs["file.version"]
            except KeyError:
                return self.__file_version
        return self.__file_version

    @file_version.setter
    def file_version(self, value):
        """set file version while validating input"""
        if not isinstance(value, str):
            msg = f"Input file version must be a string not {type(value)}"
            self.logger.error(msg)
            raise ValueError(msg)
        if value not in acceptable_file_versions:
            msg = f"Input file.version is not valid, must be {acceptable_file_versions}"
            self.logger.error(msg)
            raise ValueError(msg)
        self.__file_version = value
        self._set_default_groups()

        if self.h5_is_read():
            self.__hdf5_obj.attrs["file.version"] = value

    @property
    def software_name(self):
        """software name that wrote the file"""
        if self.h5_is_read():
            return self.__hdf5_obj.attrs["mth5.software.name"]
        return "mth5"

    @property
    def data_level(self):
        """data level"""
        if self.h5_is_read():
            try:
                return self.__hdf5_obj.attrs["data_level"]
            except KeyError:
                return self.__data_level
        else:
            return self.__data_level

    @data_level.setter
    def data_level(self, value):
        """set data level while validating input"""
        if not isinstance(value, int):
            msg = f"Input file type must be an integer not {type(value)}"
            self.logger.error(msg)
            raise ValueError(msg)
        if value not in acceptable_data_levels:
            msg = f"Input data_level is not valid, must be {acceptable_data_levels}"
            self.logger.error(msg)
            raise ValueError(msg)
        self.__data_level = value

        if self.h5_is_read():
            self.__hdf5_obj.attrs["data_level"] = value

    def _set_default_groups(self):
        """get the default groups based on file version"""

        if self.file_version in ["0.1.0"]:

            self._default_root_name = "Survey"
            self._default_subgroup_names = [
                "Stations",
                "Reports",
                "Filters",
                "Standards",
            ]

            self._root_path = "/Survey"
        elif self.file_version in ["0.2.0"]:
            self._default_root_name = "Experiment"
            self._default_subgroup_names = [
                "Surveys",
                "Reports",
                "Standards",
            ]

            self._root_path = "/Experiment"

    @property
    def experiment_group(self):
        """Convenience property for /Experiment group"""
        if self.h5_is_read():
            if self.file_version in ["0.2.0"]:
                return groups.ExperimentGroup(
                    self.__hdf5_obj[f"{self._root_path}"],
                    **self.dataset_options,
                )
            else:
                self.logger.info(
                    f"File version {self.file_version} does not have an Experiment Group"
                )
                return None
        self.logger.info("File is closed cannot access /Experiment")
        return None

    @property
    def survey_group(self):
        """Convenience property for /Survey group"""
        if self.file_version in ["0.1.0"]:
            if self.h5_is_read():
                return groups.SurveyGroup(
                    self.__hdf5_obj[f"{self._root_path}"],
                    **self.dataset_options,
                )
            self.logger.info("File is closed cannot access /Survey")
            return None
        elif self.file_version in ["0.2.0"]:
            self.logger.info(
                f"File version {self.file_version} does not have a survey_group, try surveys_group"
            )

    @property
    def surveys_group(self):
        """Convenience property for /Surveys group"""
        if self.file_version in ["0.1.0"]:
            self.logger.info(
                f"File version {self.file_version} does not have a survey_group, try surveys_group"
            )
        elif self.file_version in ["0.2.0"]:
            if self.h5_is_read():
                return groups.MasterSurveyGroup(
                    self.__hdf5_obj[f"{self._root_path}/Surveys"],
                    **self.dataset_options,
                )
            self.logger.info("File is closed cannot access /Surveys")
            return None

    @property
    def reports_group(self):
        """Convenience property for /Survey/Reports group"""
        if self.h5_is_read():
            return groups.ReportsGroup(
                self.__hdf5_obj[f"{self._root_path}/Reports"],
                **self.dataset_options,
            )
        self.logger.info("File is closed cannot access /Reports")
        return None

    @property
    def filters_group(self):
        """Convenience property for /Survey/Filters group"""
        if self.h5_is_read():
            if self.file_version in ["0.1.0"]:
                return groups.FiltersGroup(
                    self.__hdf5_obj[f"{self._root_path}/Filters"],
                    **self.dataset_options,
                )
            else:
                self.logger.info(
                    "File version 0.2.0 does not have a FiltersGroup at the experiment level"
                )
                return None
        self.logger.info("File is closed cannot access /Filters")
        return None

    @property
    def standards_group(self):
        """Convenience property for /Standards group"""
        if self.h5_is_read():
            return groups.StandardsGroup(
                self.__hdf5_obj[f"{self._root_path}/Standards"],
                **self.dataset_options,
            )
        self.logger.info("File is closed cannot access /Standards")
        return None

    @property
    def stations_group(self):
        """Convenience property for /Survey/Stations group"""
        if self.h5_is_read():
            if self.file_version not in ["0.1.0"]:
                self.logger.info(
                    f"File version {self.file_version} does not have a Stations. "
                    "try surveys_group."
                )
                return None
            return groups.MasterStationGroup(
                self.__hdf5_obj[f"{self._root_path}/Stations"],
                **self.dataset_options,
            )
        self.logger.info("File is closed cannot access /Stations")
        return None

    @property
    def station_list(self):
        """list of existing stations names"""
        if self.file_version in ["0.1.0"]:
            return self.stations_group.groups_list
        elif self.file_version in ["0.2.0"]:
            station_list = []
            for survey in self.surveys_group.groups_list:
                sg = self.surveys_group.get_survey(survey)
                station_list += sg.stations_group.groups_list
            return station_list

    def open_mth5(self, filename=None, mode="a"):
        """
        open an mth5 file

        :return: Survey Group
        :type: groups.SurveyGroup

        :Example:

        >>> from mth5 import mth5
        >>> mth5_object = mth5.MTH5(file_version='0.1.0')
        >>> survey_object = mth5_object.open_mth5('Test.mth5', 'w')

        >>> from mth5 import mth5
        >>> mth5_object = mth5.MTH5()
        >>> survey_object = mth5_object.open_mth5('Test.mth5', 'w')
        >>> mth5_object.file_version
        '0.2.0'


        """
        if filename is not None:
            self.__filename = filename
        if not isinstance(self.__filename, Path):
            self.__filename = Path(filename)
        if self.__filename.exists():
            if mode in ["w"]:
                self.logger.warning(
                    f"{self.__filename.name} will be overwritten in 'w' mode"
                )
                try:
                    self._initialize_file(mode)
                except OSError as error:
                    msg = (
                        f"{error}. Need to close any references to {self.__filename} first. "
                        + "Then reopen the file in the preferred mode"
                    )
                    self.logger.exception(msg)
            elif mode in ["a", "w-", "x", "r+"]:
                self.__hdf5_obj = h5py.File(self.__filename, mode=mode)
                self._set_default_groups()
                if not self.validate_file():
                    msg = "Input file is not a valid MTH5 file"
                    self.logger.error(msg)
                    raise MTH5Error(msg)
            elif mode in ["r"]:
                self.__hdf5_obj = h5py.File(self.__filename, mode=mode)
                self._set_default_groups()
                self.validate_file()
            else:
                msg = "mode {0} is not understood".format(mode)
                self.logger.error(msg)
                raise MTH5Error(msg)
        else:
            if mode in ["a", "w", "w-", "x"]:
                self._initialize_file(mode=mode)
            else:
                msg = "Cannot open new file in mode {0} ".format(mode)
                self.logger.error(msg)
                raise MTH5Error(msg)
        # TODO need to add a validation step to check for version and legit file
        if not "channel_summary" in self.__hdf5_obj[self._root_path].keys():
            self._initialize_summary()

    def _initialize_file(self, mode="w"):
        """
        Initialize the default groups for the file

        :return: Survey Group
        :rtype: groups.SurveyGroup

        """
        # open an hdf5 file
        self.__hdf5_obj = h5py.File(self.__filename, mode)

        # write general metadata
        self.__hdf5_obj.attrs.update(self.file_attributes)

        # create the default group
        root = self.__hdf5_obj.create_group(self._default_root_name)
        if self._default_root_name == "Survey":
            root_group = groups.SurveyGroup(root)
            root_group.write_metadata()
        for group_name in self._default_subgroup_names:
            try:
                self.__hdf5_obj.create_group(
                    f"{self._default_root_name}/{group_name}"
                )
            except ValueError:
                pass
            m5_grp = getattr(self, f"{group_name.lower()}_group")
            m5_grp.initialize_group()
        self._initialize_summary()

        self.logger.info(
            f"Initialized MTH5 {self.file_version} file {self.filename} in mode {mode}"
        )

    def _initialize_summary(self):
        try:
            # initiate channel and tf summary datasets
            self.__hdf5_obj[self._default_root_name].create_dataset(
                "channel_summary",
                shape=(1,),
                maxshape=(None,),
                dtype=CHANNEL_DTYPE,
                **self.dataset_options,
            )
        except ValueError:
            pass
        try:
            self.__hdf5_obj[self._default_root_name].create_dataset(
                "tf_summary",
                shape=(1,),
                maxshape=(None,),
                dtype=TF_DTYPE,
                **self.dataset_options,
            )
        except ValueError:
            pass

    def validate_file(self):
        """
        Validate an open mth5 file

        will test the attribute values and group names

        :return: Boolean [ True = valid, False = not valid]
        :rtype: Boolean

        """

        if self.h5_is_read():
            if self.file_type not in acceptable_file_types:
                msg = f"Unacceptable file type {self.file_type}"
                self.logger.error(msg)
                return False
            if self.file_version not in acceptable_file_versions:
                msg = f"Unacceptable file version {self.file_version}"
                self.logger.error(msg)
                return False
            if self.data_level not in acceptable_data_levels:
                msg = f"Unacceptable data_level {self.data_level}"
                self.logger.error(msg)
                return False
            if self.file_version in ["0.1.0"]:
                for gr in self.survey_group.groups_list:
                    if "summary" in gr:
                        continue
                    if gr not in self._default_subgroup_names:
                        msg = f"Unacceptable group {gr}"
                        self.logger.error(msg)
                        return False
            elif self.file_version in ["0.2.0"]:
                for gr in self.experiment_group.groups_list:
                    if "summary" in gr:
                        continue
                    if gr not in self._default_subgroup_names:
                        msg = f"Unacceptable group {gr}"
                        self.logger.error(msg)
                        return False
            return True
        self.logger.warning("HDF5 file is not open")
        return False

    def close_mth5(self):
        """
        close mth5 file to make sure everything is flushed to the file
        """

        try:
            # update summay tables
            if self.h5_is_write():
                self.channel_summary.summarize()
                self.tf_summary.summarize()
            self.logger.info(f"Flushing and closing {str(self.filename)}")
            self.__hdf5_obj.flush()
            self.__hdf5_obj.close()
        except (AttributeError, ValueError):
            helpers.close_open_files()

    def h5_is_write(self):
        """
        check to see if the hdf5 file is open and writeable
        """
        if isinstance(self.__hdf5_obj, h5py.File):
            try:
                if "w" in self.__hdf5_obj.mode or "+" in self.__hdf5_obj.mode:
                    return True
                return False
            except ValueError:
                return False
        return False

    def h5_is_read(self):
        """
        check to see if the hdf5 file is open and readable

        :return: True if readable, False if not
        :rtype: Boolean

        """

        if isinstance(self.__hdf5_obj, h5py.File):
            try:
                if self.__hdf5_obj.mode in ["r", "r+", "a", "w", "w-", "x"]:
                    return True
                return False
            except ValueError:
                return False
        return False

    def has_group(self, group_name):
        """
        Check to see if the group name exists
        """
        if self.h5_is_read():

            def has_name(name):
                if group_name == name:
                    return True

            if self.__hdf5_obj.visit(has_name):
                return True
            return False

    def _make_h5_path(
        self, survey=None, station=None, run=None, channel=None, tf_id=None
    ):
        """
        create an h5 path from inputs
        """

        if self.file_version == "0.1.0":
            h5_path = self._root_path

        elif self.file_version == "0.2.0":
            if survey is None:
                raise ValueError("Survey must be input for file type 0.2.0")
            else:
                survey = helpers.validate_name(survey)
            h5_path = f"{self._root_path}/Surveys/{survey}"

        if station is not None:
            station = helpers.validate_name(station)
            h5_path += f"/Stations/{station}"

            if tf_id is not None:
                tf_id = helpers.validate_name(tf_id)
                h5_path += f"/Transfer_Functions/{tf_id}"
            elif run is not None:
                run = helpers.validate_name(run)
                h5_path += f"/{run}"
                if channel is not None:
                    channel = helpers.validate_name(channel)
                    h5_path += f"/{channel}"

        return h5_path

    def from_reference(self, h5_reference):
        """
        Get an HDF5 group, dataset, etc from a reference

        :param h5_reference: DESCRIPTION
        :type h5_reference: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """
        ref_dict = {
            "survey": groups.SurveyGroup,
            "station": groups.StationGroup,
            "run": groups.RunGroup,
            "electric": groups.ElectricDataset,
            "magnetic": groups.MagneticDataset,
            "auxiliary": groups.AuxiliaryDataset,
            "transferfunction": groups.TransferFunctionGroup,
        }

        # in the future should allow this to return the proper container.
        referenced = self.__hdf5_obj[h5_reference]
        mth5_type = referenced.attrs["mth5_type"].lower()

        try:
            if mth5_type == "transferfunction":
                group = ref_dict[mth5_type](referenced)
                return group.to_tf_object()
            return ref_dict[mth5_type](referenced)
        except KeyError:
            self.logger.info(
                f"Could not identify the MTH5 type {mth5_type}, "
                + "returning h5 group."
            )
            return referenced

    def to_experiment(self):
        """
        Create an :class:`mt_metadata.timeseries.Experiment` object from the
        metadata contained in the MTH5 file.

        :returns: :class:`mt_metadata.timeseries.Experiment`

        """

        if self.h5_is_read():
            if self.file_version in ["0.1.0"]:
                experiment = Experiment()
                experiment.surveys.append(self.survey_group.metadata)
            elif self.file_version in ["0.2.0"]:
                experiment = self.experiment_group.metadata
            return experiment

    def from_experiment(self, experiment, survey_index=0, update=False):
        """
        Fill out an MTH5 from a :class:`mt_metadata.timeseries.Experiment` object
        given a survey_id

        :param experiment: Experiment metadata
        :type experiment: :class:`mt_metadata.timeseries.Experiment`
        :param survey_index: Index of the survey to write
        :type survey_index: int, defaults to 0

        """
        if self.h5_is_write():
            if self.file_version in ["0.1.0"]:
                sg = self.survey_group
                sg.metadata.from_dict(
                    experiment.surveys[survey_index].to_dict()
                )
                sg.write_metadata()
                for station in experiment.surveys[0].stations:
                    mt_station = self.add_station(
                        station.id, station_metadata=station
                    )
                    if update:
                        mt_station.metadata.update(station)
                        mt_station.write_metadata()
                    for run in station.runs:
                        mt_run = mt_station.add_run(run.id, run_metadata=run)
                        if update:
                            mt_run.metadata.update(run)
                            mt_run.write_metadata()
                        for channel in run.channels:
                            mt_ch = mt_run.add_channel(
                                channel.component,
                                channel.type,
                                None,
                                channel_metadata=channel,
                            )
                            if update:
                                mt_ch.metadata.update(channel)
                                mt_ch.write_metadata()
                        ### need to update from input metadata for time period
                        ### and channels, runs, stations.
                        mt_run.update_run_metadata()
                    mt_station.update_station_metadata()
                sg.update_survey_metadata()
                for k, v in experiment.surveys[0].filters.items():
                    self.filters_group.add_filter(v)
            elif self.file_version in ["0.2.0"]:

                for survey in experiment.surveys:
                    sg = self.add_survey(survey.id, survey_metadata=survey)

                    for station in survey.stations:
                        mt_station = self.add_station(
                            station.id,
                            station_metadata=station,
                            survey=sg.metadata.id,
                        )
                        if update:
                            mt_station.metadata.update(station)
                            mt_station.write_metadata()
                        for run in station.runs:
                            mt_run = mt_station.add_run(
                                run.id, run_metadata=run
                            )
                            if update:
                                mt_run.metadata.update(run)
                                mt_run.write_metadata()
                            for channel in run.channels:
                                mt_ch = mt_run.add_channel(
                                    channel.component,
                                    channel.type,
                                    None,
                                    channel_metadata=channel,
                                )
                                if update:
                                    mt_ch.metadata.update(channel)
                                    mt_ch.write_metadata()

                            mt_run.update_run_metadata()
                        mt_station.update_station_metadata()
                    sg.update_survey_metadata()
                    for k, v in survey.filters.items():
                        sg.filters_group.add_filter(v)

    @property
    def channel_summary(self):
        """return a dataframe of channels"""

        return ChannelSummaryTable(
            self.__hdf5_obj[f"{self._root_path}/channel_summary"]
        )

    @property
    def tf_summary(self):
        """return a dataframe of channels"""

        return TFSummaryTable(self.__hdf5_obj[f"{self._root_path}/tf_summary"])

    def add_survey(self, survey_name, survey_metadata=None):
        """
        Add a survey with metadata if given with the path:
            ``/Experiment/Surveys/survey_name``

        If the survey already exists, will return that survey and nothing
        is added.

        :param survey_name: Name of the survey, should be the same as
                             metadata.id
        :type survey_name: string
        :param survey_metadata: survey metadata container, defaults to None
        :type survey_metadata: :class:`mth5.metadata.survey`, optional
        :return: A convenience class for the added survey
        :rtype: :class:`mth5_groups.SurveyGroup`

        :Example: ::

            >>> from mth5 import mth5
            >>> mth5_obj = mth5.MTH5()
            >>> mth5_obj.open_mth5(r"/test.mth5", mode='a')
            >>> # one option
            >>> new_survey = mth5_obj.add_survey('MT001')
            >>> # another option
            >>> new_station = mth5_obj.experiment_group.surveys_group.add_survey('MT001')

        """
        return self.surveys_group.add_survey(
            survey_name, survey_metadata=survey_metadata
        )

    def get_survey(self, survey_name):
        """
        Get a survey with the same name as survey_name

        :param survey_name: existing survey name
        :type survey_name: string
        :return: convenience survey class
        :rtype: :class:`mth5.mth5_groups.surveyGroup`
        :raises MTH5Error:  if the survey name is not found.

        :Example:

        >>> from mth5 import mth5
        >>> mth5_obj = mth5.MTH5()
        >>> mth5_obj.open_mth5(r"/test.mth5", mode='a')
        >>> # one option
        >>> existing_survey = mth5_obj.get_survey('MT001')
        >>> # another option
        >>> existing_staiton = mth5_obj.experiment_group.surveys_group.get_survey('MT001')
        MTH5Error: MT001 does not exist, check groups_list for existing names

        """

        survey_path = self._make_h5_path(survey=survey_name)
        try:
            return groups.SurveyGroup(
                self.__hdf5_obj[survey_path],
                **self.dataset_options,
            )
        except KeyError:
            msg = (
                f"{survey_path} does not exist, check survey_list for "
                "existing names."
            )
            self.logger.warning(msg)
            raise MTH5Error(msg)

    def remove_survey(self, survey_name):
        """
        Remove a survey from the file.

        .. note:: Deleting a survey is not as simple as del(survey).  In HDF5
              this does not free up memory, it simply removes the reference
              to that survey.  The common way to get around this is to
              copy what you want into a new file, or overwrite the survey.

        :param survey_name: existing survey name
        :type survey_name: string

        :Example: ::

            >>> from mth5 import mth5
            >>> mth5_obj = mth5.MTH5()
            >>> mth5_obj.open_mth5(r"/test.mth5", mode='a')
            >>> # one option
            >>> mth5_obj.remove_survey('MT001')
            >>> # another option
            >>> mth5_obj.experiment_group.surveys_group.remove_survey('MT001')

        """

        survey_path = self._make_h5_path(survey=survey_name)
        try:
            del self.__hdf5_obj[f"{survey_path}"]
            self.logger.info(
                "Deleting a survey does not reduce the HDF5"
                "file size it simply remove the reference. If "
                "file size reduction is your goal, simply copy"
                " what you want into another file."
            )
        except KeyError:
            msg = (
                f"{survey_path} does not exist, "
                "check station_list for existing names"
            )
            self.logger.warning(msg)
            raise MTH5Error(msg)

    def add_station(self, station_name, station_metadata=None, survey=None):
        """
        Convenience function to add a station using
        ``mth5.stations_group.add_station``


        Add a station with metadata if given with the path [v0.1.0]:
            ``/Survey/Stations/station_name``

        Add a station with metadata if given with the path [v0.2.0]:
            ``Experiment/Surveys/survey/Stations/station_name``

        If the station already exists, will return that station and nothing
        is added.

        :param station_name: Name of the station, should be the same as
                             metadata.archive_id
        :type station_name: string
        :param station_metadata: Station metadata container, defaults to None
        :type station_metadata: :class:`mth5.metadata.Station`, optional
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string
        :return: A convenience class for the added station
        :rtype: :class:`mth5_groups.StationGroup`

        :Example:

        >>> new_staiton = mth5_obj.add_station('MT001')

        """

        if self.file_version in ["0.1.0"]:
            return self.stations_group.add_station(
                station_name, station_metadata=station_metadata
            )
        elif self.file_version in ["0.2.0"]:
            if survey is None:
                msg = "Need to input 'survey' for file version %s"
                self.logger.error(msg, self.file_version)
                raise ValueError(msg % self.file_version)
            sg = self.get_survey(survey)
            return sg.stations_group.add_station(
                station_name, station_metadata=station_metadata
            )

    def get_station(self, station_name, survey=None):
        """
        Convenience function to get a station using

        Get a station with the same name as station_name

        :param station_name: existing station name
        :type station_name: string
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string
        :return: convenience station class
        :rtype: :class:`mth5.mth5_groups.StationGroup`
        :raises MTH5Error:  if the station name is not found.

        :Example:

        >>> existing_staiton = mth5_obj.get_station('MT001')
        MTH5Error: MT001 does not exist, check station_list for existing names

        """

        station_path = self._make_h5_path(survey=survey, station=station_name)
        try:
            return groups.StationGroup(self.__hdf5_obj[station_path])
        except KeyError:
            raise MTH5Error(f"Could not find station {station_name}")

    def remove_station(self, station_name, survey=None):
        """
        Convenience function to remove a station using

        Remove a station from the file.

        .. note:: Deleting a station is not as simple as del(station).  In HDF5
              this does not free up memory, it simply removes the reference
              to that station.  The common way to get around this is to
              copy what you want into a new file, or overwrite the station.

        :param station_name: existing station name
        :type station_name: string
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string

        :Example:

        >>> mth5_obj.remove_station('MT001')

        """
        station_name = helpers.validate_name(station_name)
        if self.file_version in ["0.1.0"]:
            return self.stations_group.remove_station(station_name)
        elif self.file_version in ["0.2.0"]:
            if survey is None:
                msg = "Need to input 'survey' for file version %s"
                self.logger.error(msg, self.file_version)
                raise ValueError(msg % self.file_version)
            survey = helpers.validate_name(survey)
            sg = self.get_survey(survey)
            return sg.stations_group.remove_station(station_name)

    def add_run(self, station_name, run_name, run_metadata=None, survey=None):
        """
        Convenience function to add a run using

        Add a run to a given station.

        :param run_name: run name, should be archive_id{a-z}
        :type run_name: string
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string
        :param metadata: metadata container, defaults to None
        :type metadata: :class:`mth5.metadata.Station`, optional

        :example:

        >>> new_run = mth5_obj.add_run('MT001', 'MT001a')


        .. todo:: auto fill run name if none is given.

        .. todo:: add ability to add a run with data.


        """

        return self.get_station(station_name, survey=survey).add_run(
            run_name, run_metadata=run_metadata
        )

    def get_run(self, station_name, run_name, survey=None):
        """
        Convenience function to get a run using
        ``mth5.stations_group.get_station(station_name).get_run()``

        get a run from run name for a given station

        :param station_name: existing station name
        :type station_name: string
        :param run_name: existing run name
        :type run_name: string
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string
        :return: Run object
        :rtype: :class:`mth5.mth5_groups.RunGroup`

        :Example:

        >>> existing_run = mth5_obj.get_run('MT001', 'MT001a')

        """

        run_path = self._make_h5_path(
            survey=survey, station=station_name, run=run_name
        )
        try:
            return groups.RunGroup(self.__hdf5_obj[run_path])
        except KeyError:
            raise MTH5Error(f"Could not find {run_path}")

    def remove_run(self, station_name, run_name, survey=None):
        """
        Remove a run from the station.

        .. note:: Deleting a run is not as simple as del(run).  In HDF5
              this does not free up memory, it simply removes the reference
              to that station.  The common way to get around this is to
              copy what you want into a new file, or overwrite the run.

        :param station_name: existing station name
        :type station_name: string
        :param run_name: existing run name
        :type run_name: string
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string

        :Example:

        >>> mth5_obj.remove_station('MT001', 'MT001a')

        """

        return self.get_station(station_name, survey=survey).remove_run(
            run_name
        )

    def add_channel(
        self,
        station_name,
        run_name,
        channel_name,
        channel_type,
        data,
        channel_dtype="int32",
        max_shape=(None,),
        chunks=True,
        channel_metadata=None,
        survey=None,
    ):
        """
        Convenience function to add a channel using
        ``mth5.stations_group.get_station().get_run().add_channel()``

        add a channel to a given run for a given station

        :param station_name: existing station name
        :type station_name: string
        :param run_name: existing run name
        :type run_name: string
        :param channel_name: name of the channel
        :type channel_name: string
        :param channel_type: [ electric | magnetic | auxiliary ]
        :type channel_type: string
        :raises MTH5Error: If channel type is not correct

        :param channel_metadata: metadata container, defaults to None
        :type channel_metadata: [ :class:`mth5.metadata.Electric` |
                                 :class:`mth5.metadata.Magnetic` |
                                 :class:`mth5.metadata.Auxiliary` ], optional
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string
        :return: Channel container
        :rtype: [ :class:`mth5.mth5_groups.ElectricDatset` |
                 :class:`mth5.mth5_groups.MagneticDatset` |
                 :class:`mth5.mth5_groups.AuxiliaryDatset` ]

        :Example:

        >>> new_channel = mth5_obj.add_channel('MT001', 'MT001a''Ex',
        >>> ...                                'electric', None)
        >>> new_channel
        Channel Electric:
        -------------------
                        component:        None
                data type:        electric
                data format:      float32
                data shape:       (1,)
                start:            1980-01-01T00:00:00+00:00
                end:              1980-01-01T00:00:00+00:00
                sample rate:      None


        """

        return self.get_run(station_name, run_name, survey=survey).add_channel(
            channel_name,
            channel_type,
            data,
            channel_metadata=channel_metadata,
            channel_dtype=channel_dtype,
            max_shape=max_shape,
            chunks=chunks,
            **self.dataset_options,
        )

    def get_channel(self, station_name, run_name, channel_name, survey=None):
        """
        Convenience function to get a channel using
        ``mth5.stations_group.get_station().get_run().get_channel()``

        Get a channel from an existing name.  Returns the appropriate
        container.

        :param station_name: existing station name
        :type station_name: string
        :param run_name: existing run name
        :type run_name: string
        :param channel_name: name of the channel
        :type channel_name: string
        :return: Channel container
        :rtype: [ :class:`mth5.mth5_groups.ElectricDatset` |
                  :class:`mth5.mth5_groups.MagneticDatset` |
                  :class:`mth5.mth5_groups.AuxiliaryDatset` ]
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string
        :raises MTH5Error:  If no channel is found

        :Example:

        >>> existing_channel = mth5_obj.get_channel(station_name,
        >>> ...                                     run_name,
        >>> ...                                     channel_name)
        >>> existing_channel
        Channel Electric:
        -------------------
                        component:        Ex
                data type:        electric
                data format:      float32
                data shape:       (4096,)
                start:            1980-01-01T00:00:00+00:00
                end:              1980-01-01T00:00:01+00:00
                sample rate:      4096

        """
        run_path = self._make_h5_path(
            survey=survey, station=station_name, run=run_name
        )
        rg = groups.RunGroup(self.__hdf5_obj[run_path])
        try:
            return rg.get_channel(helpers.validate_name(channel_name))
        except (AttributeError, KeyError):
            raise MTH5Error(
                f"Could not find channel, {run_path}/{channel_name}"
            )

    def remove_channel(self, station_name, run_name, channel_name, survey=None):
        """
        Convenience function to remove a channel using
        ``mth5.stations_group.get_station().get_run().remove_channel()``

        Remove a channel from a given run and station.

        .. note:: Deleting a channel is not as simple as del(channel).  In HDF5
              this does not free up memory, it simply removes the reference
              to that channel.  The common way to get around this is to
              copy what you want into a new file, or overwrite the channel.

        :param station_name: existing station name
        :type station_name: string
        :param run_name: existing run name
        :type run_name: string
        :param channel_name: existing station name
        :type channel_name: string
        :param survey: existing survey name, needed for file version >= 0.2.0
        :type survey: string

        :Example:

        >>> mth5_obj.remove_channel('MT001', 'MT001a', 'Ex')

        """
        station_name = helpers.validate_name(station_name)
        run_name = helpers.validate_name(run_name)
        channel_name = helpers.validate_name(channel_name)
        return (
            self.get_station(station_name, survey=survey)
            .get_run(run_name)
            .remove_channel(channel_name)
        )

    def add_transfer_function(self, tf_object):
        """
        Add a transfer function
        :param tf_object: DESCRIPTION
        :type tf_object: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if not isinstance(tf_object, TF):
            msg = "Input must be a TF object not %s"
            self.logger.error(msg, type(tf_object))
            raise ValueError(msg % type(tf_object))
        if self.file_version == "0.2.0":
            try:
                # need to check survey metadata to make sure it matches,
                # if it doesn't need to make a new survey group so that
                # when a TF is pulled it gets the proper survey metadata.
                # this should eventually search over each unknonw survey
                # for matching metadata so there aren't 100 groups
                survey_group = self.get_survey(tf_object.survey_metadata.id)
                if tf_object.survey_metadata.id == "unknown_survey":
                    for sg_id in self.surveys_group.groups_list:
                        if "unknown_survey" in sg_id:
                            match = True
                            survey_group = self.get_survey(sg_id)
                            sg_dict = survey_group.metadata.to_dict(
                                single=True, required=False
                            )
                            for (
                                key,
                                value,
                            ) in tf_object.survey_metadata.to_dict(
                                single=True
                            ).items():
                                if key in [
                                    "hdf5_reference",
                                    "mth5_type",
                                    "id",
                                ]:
                                    continue
                                if sg_dict[key] != value:
                                    match = False
                                    break
                            if match:
                                break
                    # create a new survey group with a new id, this is likely
                    # not the best way to do this, it should be strongly
                    # encouraged that the user assigne a survey id.
                    if not match:
                        count = 1
                        survey_id = f"unknown_survey_{count:03}"
                        while survey_id in self.surveys_group.groups_list:
                            count += 1
                            survey_id = f"unknown_survey_{count:03}"
                        tf_object.survey_metadata.id = survey_id
                        survey_group = self.add_survey(
                            tf_object.survey_metadata.id,
                            survey_metadata=tf_object.survey_metadata,
                        )
            except MTH5Error:
                survey_group = self.add_survey(
                    tf_object.survey_metadata.id,
                    survey_metadata=tf_object.survey_metadata,
                )
        else:
            survey_group = self.survey_group
            # might need a better test here
            if survey_group.metadata.id is None:
                survey_group.metadata.update(tf_object.survey_metadata)
                survey_group.write_metadata()
        try:
            station_group = survey_group.stations_group.get_station(
                tf_object.station_metadata.id
            )
            station_group.metadata.update(tf_object.to_ts_station_metadata())
            station_group.write_metadata()
        except MTH5Error:
            station_group = survey_group.stations_group.add_station(
                tf_object.station_metadata.id,
                station_metadata=tf_object.to_ts_station_metadata(),
            )
        ## need to check for runs and channels
        for (
            run_id
        ) in tf_object.station_metadata.transfer_function.runs_processed:
            if run_id in ["", None, "None"]:
                continue
            try:
                run_group = station_group.get_run(run_id)
            except MTH5Error:
                run = tf_object.station_metadata.get_run(run_id)
                run_group = station_group.add_run(run_id, run_metadata=run)

                for ch in run.channels:
                    try:
                        ch_dataset = run_group.get_channel(ch.component)
                    except MTH5Error:
                        ch_dataset = run_group.add_channel(
                            ch.component, ch.type, None, channel_metadata=ch
                        )
        try:
            tf_group = (
                station_group.transfer_functions_group.add_transfer_function(
                    tf_object.station, tf_object=tf_object
                )
            )
        except (OSError, RuntimeError, ValueError):
            msg = f"TF {tf_object.station} already exists, returning existing group."
            self.logger.debug(msg)
            tf_group = (
                station_group.transfer_functions_group.get_transfer_function(
                    tf_object.station
                )
            )
        return tf_group

    def get_transfer_function(self, station_id, tf_id, survey=None):
        """
        Get a transfer function

        :param survey_id: DESCRIPTION
        :type survey_id: TYPE
        :param station_id: DESCRIPTION
        :type station_id: TYPE
        :param tf_id: DESCRIPTION
        :type tf_id: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        try:
            tf_df = self.tf_summary.to_dataframe()
            ref = (
                tf_df.loc[
                    (tf_df.station == station_id)
                    & (tf_df.tf_id == tf_id)
                    & (tf_df.survey == "unknown_survey")
                ]
                .iloc[0]
                .hdf5_reference
            )
            return self.from_reference(ref)
        except IndexError:
            tf_path = self._make_h5_path(
                survey=survey, station=station_id, tf_id=tf_id
            )
            try:
                tg = groups.TransferFunctionGroup(self.__hdf5_obj[tf_path])
                return tg.to_tf_object()
            except KeyError:
                raise MTH5Error(f"Could not find {tf_path}")

    def remove_transfer_function(self, station_id, tf_id, survey=None):
        """
        remove a transfer function

        :param survey_id: DESCRIPTION
        :type survey_id: TYPE
        :param station_id: DESCRIPTION
        :type station_id: TYPE
        :param tf_id: DESCRIPTION
        :type tf_id: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        station_group = self.get_station(station_id, survey=survey)

        station_group.transfer_functions_group.remove_transfer_function(tf_id)



0  # -*- coding: utf-8 -*-
"""
Created on Fri Aug 19 16:39:30 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
from pathlib import Path
from collections import OrderedDict

import numpy as np
from mth5.io.phoenix import open_phoenix

# =============================================================================


@unittest.skipIf(
    "peacock" not in str(Path(__file__).as_posix()),
    "Only local files, cannot test in GitActions",
)
class TestReadPhoenixContinuous(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.phx_obj = open_phoenix(
            r"c:\Users\jpeacock\OneDrive - DOI\mt\phoenix_example_data\Sample Data\10128_2021-04-27-032436\0\10128_608783F4_0_00000001.td_150"
        )

        self.data = self.phx_obj.read_sequence()

        self.original = open_phoenix(
            r"c:\Users\jpeacock\OneDrive - DOI\mt\phoenix_example_data\Sample Data\10128_2021-04-27-032436\0\10128_608783F4_0_00000001.td_150"
        )
        self.original_data = self.original.read()

        self.maxDiff = None

    def test_readers_match(self):
        self.assertTrue(
            np.allclose(
                self.data[0 : self.original_data[0].size],
                self.original_data[0],
            )
        )

    def test_attributes(self):
        true_dict = {
            "ad_plus_minus_range": 5.0,
            "attenuator_gain": 1.0,
            "base_dir": Path(
                "c:/Users/jpeacock/OneDrive - DOI/mt/phoenix_example_data/Sample Data/10128_2021-04-27-032436/0"
            ),
            "base_path": Path(
                "c:/Users/jpeacock/OneDrive - DOI/mt/phoenix_example_data/Sample Data/10128_2021-04-27-032436/0/10128_608783F4_0_00000001.td_150"
            ),
            "battery_voltage_v": 12.48,
            "board_model_main": "BCM01",
            "board_model_revision": "",
            "bytes_per_sample": 4,
            "ch_board_model": "BCM01-I",
            "ch_board_serial": 200803,
            "ch_firmware": 65567,
            "channel_id": 0,
            "channel_main_gain": 4.0,
            "channel_map": {
                0: "hx",
                1: "hy",
                2: "hz",
                3: "ex",
                4: "ey",
                5: "h1",
                6: "h2",
                7: "h3",
            },
            "channel_type": "H",
            "data_footer": 0,
            "decimation_node_id": 2,
            "detected_channel_type": "H",
            "file_extension": ".td_150",
            "file_name": "10128_608783F4_0_00000001.td_150",
            "file_sequence": 2,
            "file_size": 215528,
            "file_type": 2,
            "file_version": 2,
            "frag_period": 360,
            "frame_rollover_count": 0,
            "frame_size": 64,
            "frame_size_bytes": 64,
            "future1": 33,
            "future2": 0,
            "gps_elevation": 181.12939453125,
            "gps_horizontal_accuracy": 19.899,
            "gps_lat": 43.696022033691406,
            "gps_long": -79.39376831054688,
            "gps_vertical_accuracy": 22.873,
            "hardware_configuration": (4, 3, 0, 0, 0, 9, 128, 0),
            "header_length": 128,
            "instrument_id": "10128",
            "instrument_serial_number": "10128",
            "instrument_type": "MTU-5C",
            "intrinsic_circuitry_gain": 1.0,
            "last_seq": 2,
            "lp_frequency": 10000,
            "max_samples": 53850,
            "max_signal": 2.0414257049560547,
            "min_signal": -2.0301353931427,
            "missing_frames": 0,
            "preamp_gain": 1.0,
            "recording_id": 1619493876,
            "recording_start_time": "2021-04-26T20:24:36+00:00",
            "report_hw_sat": False,
            "sample_rate": 150,
            "sample_rate_base": 150,
            "sample_rate_exp": 0,
            "saturated_frames": 0,
            "segment_end_time": "2021-04-26T20:36:35+00:00",
            "segment_start_time": "2021-04-26T20:30:36+00:00",
            "seq": 1,
            "sequence_list": [
                Path(
                    "c:/Users/jpeacock/OneDrive - DOI/mt/phoenix_example_data/Sample Data/10128_2021-04-27-032436/0/10128_608783F4_0_00000001.td_150"
                ),
                Path(
                    "c:/Users/jpeacock/OneDrive - DOI/mt/phoenix_example_data/Sample Data/10128_2021-04-27-032436/0/10128_608783F4_0_00000002.td_150"
                ),
            ],
            "subheader": {},
            "timing_flags": 55,
            "timing_sat_count": 6,
            "timing_stability": 508,
            "timing_status": (55, 6, 508),
            "total_circuitry_gain": 4.0,
            "total_selectable_gain": 4.0,
        }

        for key, original_value in true_dict.items():
            new_value = getattr(self.phx_obj, key)
            with self.subTest(f"test {key}"):
                if isinstance(original_value, (list)):
                    self.assertListEqual(original_value, new_value)
                elif isinstance(original_value, float):
                    self.assertAlmostEqual(original_value, new_value)
                else:
                    self.assertEqual(original_value, new_value)

    def test_to_channel_ts(self):
        ch_ts = self.phx_obj.to_channel_ts()

        with self.subTest("Channel metadata"):
            ch_metadata = OrderedDict(
                [
                    ("channel_number", 0),
                    ("component", "hx"),
                    ("data_quality.rating.value", 0),
                    ("filter.applied", [False]),
                    ("filter.name", []),
                    ("location.elevation", 0.0),
                    ("location.latitude", 0.0),
                    ("location.longitude", 0.0),
                    ("measurement_azimuth", 0.0),
                    ("measurement_tilt", 0.0),
                    ("sample_rate", 150.0),
                    ("sensor.id", None),
                    ("sensor.manufacturer", None),
                    ("sensor.type", None),
                    ("time_period.end", "2021-04-26T20:30:40.993333+00:00"),
                    ("time_period.start", "2021-04-26T20:24:36+00:00"),
                    ("type", "magnetic"),
                    ("units", None),
                ]
            )

            self.assertDictEqual(
                ch_ts.channel_metadata.to_dict(single=True), ch_metadata
            )

        with self.subTest("Channel Size"):
            self.assertEqual(54750, ch_ts.ts.size)



# package file
from .gps import GPS, GPSError
from .header import NIMSHeader
from .response_filters import Response
from .nims import NIMS, read_nims
from .nims_collection import NIMSCollection


__all__ = [
    "GPS",
    "GPSError",
    "NIMSHeader",
    "Response",
    "NIMS",
    "read_nims",
    "NIMSCollection",
]



# -*- coding: utf-8 -*-
"""

Created on Wed Sep 30 10:20:12 2020

:author: Jared Peacock

:license: MIT

"""
# =============================================================================
# Imports
# =============================================================================
from pathlib import Path
from mth5.timeseries import RunTS
from obspy import read as obspy_read

# =============================================================================
# read seismic file
# =============================================================================
def read_miniseed(fn):
    """
    Read a miniseed file into a :class:`mth5.timeseries.RunTS` object. Uses
    `Obspy` to read the miniseed.

    :param fn: full path to the miniseed file
    :type fn: string
    :return: RunTS object
    :rtype: :class:`mth5.timeseries.RunTS`

    """

    # obspy does not use Path objects for file names
    if isinstance(fn, Path):
        fn = fn.as_posix()
    obs_stream = obspy_read(fn)
    run_obj = RunTS()
    run_obj.from_obspy_stream(obs_stream)

    return run_obj



# -*- coding: utf-8 -*-
"""
Created on Wed Sep  7 08:38:54 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
from collections import OrderedDict

import numpy as np

from mth5.io.nims import Response

# =============================================================================


class TestNIMSResponse(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.response = Response()
        self.maxDiff = None

    def test_get_electric_hp(self):
        with self.subTest("PC"):
            pc_hp = self.response.get_electric_high_pass(hardware="pc")
            pc_dict = OrderedDict(
                [
                    ("calibration_date", "1980-01-01"),
                    ("gain", 1.0),
                    ("name", "nims_1_pole_butterworth"),
                    ("normalization_factor", 1.0),
                    ("poles", np.array([-3.333333e-05 + 0.0j])),
                    ("type", "zpk"),
                    ("units_in", "V"),
                    ("units_out", "V"),
                    ("zeros", np.array([0.0 + 0.0j])),
                ]
            )

            self.assertDictEqual(pc_dict, pc_hp.to_dict(single=True))

        with self.subTest("HP"):
            hp_hp = self.response.get_electric_high_pass(hardware="HP")
            hp_dict = OrderedDict(
                [
                    ("calibration_date", "1980-01-01"),
                    ("gain", 1.0),
                    ("name", "nims_1_pole_butterworth"),
                    ("normalization_factor", 1.0),
                    ("poles", np.array([-1.66667e-04 + 0.0j])),
                    ("type", "zpk"),
                    ("units_in", "V"),
                    ("units_out", "V"),
                    ("zeros", np.array([0.0 + 0.0j])),
                ]
            )

            self.assertDictEqual(hp_dict, hp_hp.to_dict(single=True))

    def test_dipole_filter(self):
        dp = self.response.dipole_filter(100)
        dp_dict = OrderedDict(
            [
                ("calibration_date", "1980-01-01"),
                ("gain", 100.0),
                ("name", "dipole_100.00"),
                ("type", "coefficient"),
                ("units_in", "V/m"),
                ("units_out", "V"),
            ]
        )

        self.assertDictEqual(dp_dict, dp.to_dict(single=True))

    def test_get_electric_filters(self):
        ex = self.response.get_channel_response("ex", 100)

        with self.subTest("length"):
            self.assertEqual(6, len(ex.filters_list))

        with self.subTest("units_in"):
            self.assertEqual("mV/km", ex.units_in)

        with self.subTest("units_out"):
            self.assertEqual("count", ex.units_out)

    def test_get_magnetic_filters(self):
        hx = self.response.get_channel_response("hx")

        with self.subTest("length"):
            self.assertEqual(3, len(hx.filters_list))

        with self.subTest("units_in"):
            self.assertEqual("nT", hx.units_in)

        with self.subTest("units_out"):
            self.assertEqual("count", hx.units_out)


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



"""
Import all Group objects
"""
## !!! DO NOT CHANGE ORDER !!!
from .coefficient_filter_group import CoefficientGroup
from .time_delay_filter_group import TimeDelayGroup
from .zpk_filter_group import ZPKGroup
from .fap_filter_group import FAPGroup
from .fir_filter_group import FIRGroup


__all__ = [
    "CoefficientGroup",
    "TimeDelayGroup",
    "ZPKGroup",
    "FAPGroup",
    "FIRGroup",
]



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 16 11:38:10 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
from pathlib import Path

from mth5.io.usgs_ascii import USGSascii
from mth5.timeseries import ChannelTS, RunTS

# =============================================================================


@unittest.skipIf("peacock" not in str(Path(__file__).as_posix()), "local file")
class TestUSGSAscii(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.asc = USGSascii(
            fn=r"C:\Users\jpeacock\OneDrive - DOI\mt\usgs_ascii\rgr003a_converted.asc"
        )
        self.asc.read()
        self.maxDiff = None

    def test_channel(self):
        for ch in self.asc.channel_order:
            ch_obj = getattr(self.asc, ch)

            with self.subTest("size"):
                self.assertEqual(955916, ch_obj.ts.size)

            with self.subTest("start"):
                self.assertEqual("2012-08-21T22:02:27+00:00", ch_obj.start)

            with self.subTest("end"):
                self.assertEqual("2012-08-24T16:25:25.750000+00:00", ch_obj.end)

            with self.subTest("sample rate"):
                self.assertEqual(4, ch_obj.sample_rate)

            with self.subTest("station"):
                self.assertEqual("003", ch_obj.station_metadata.id)

            with self.subTest("run"):
                self.assertEqual("rgr003a", ch_obj.run_metadata.id)

            with self.subTest("type"):
                self.assertIsInstance(ch_obj, ChannelTS)

    def test_to_run(self):
        r = self.asc.to_run_ts()

        with self.subTest("size"):
            self.assertEqual(955916, r.dataset.coords["time"].size)

        with self.subTest("start"):
            self.assertEqual("2012-08-21T22:02:27+00:00", r.start)

        with self.subTest("end"):
            self.assertEqual("2012-08-24T16:25:25.750000+00:00", r.end)

        with self.subTest("sample rate"):
            self.assertEqual(4, r.sample_rate)

        with self.subTest("station"):
            self.assertEqual("003", r.station_metadata.id)

        with self.subTest("run"):
            self.assertEqual("rgr003a", r.run_metadata.id)

        with self.subTest("channels"):
            self.assertListEqual(["hx", "hy", "hz", "ex", "ey"], r.channels)

        with self.subTest("type"):
            self.assertIsInstance(r, RunTS)


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
LEMI 424 Collection
====================

Collection of TXT files combined into runs

Created on Wed Aug 31 10:32:44 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import pandas as pd

from mth5.io.collection import Collection
from mth5.io.nims import NIMS

# =============================================================================


class NIMSCollection(Collection):
    """
    Collection of NIMS files into runs.


    .. code-block:: python

        >>> from mth5.io.nims import LEMICollection
        >>> lc = NIMSCollection(r"/path/to/single/lemi/station")
        >>> lc.station_id = "mt001"
        >>> lc.survey_id = "test_survey"
        >>> run_dict = lc.get_runs(1)


    """

    def __init__(self, file_path=None, **kwargs):
        super().__init__(file_path=file_path, **kwargs)
        self.file_ext = "bin"

        self.survey_id = "mt"

    def to_dataframe(
        self, sample_rates=[1], run_name_zeros=2, calibration_path=None
    ):
        """
        Create a data frame of each TXT file in a given directory.

        .. note:: This assumes the given directory contains a single station

        :param sample_rates: sample rate to get, will always be 1 for LEMI data
         defaults to [1]
        :type sample_rates: int or list, optional
        :param run_name_zeros: number of zeros to assing to the run name,
         defaults to 4
        :type run_name_zeros: int, optional
        :param calibration_path: path to calibration files, defaults to None
        :type calibration_path: string or Path, optional
        :return: Dataframe with information of each TXT file in the given
         directory.
        :rtype: :class:`pandas.DataFrame`

        :Example:

            >>> from mth5.io.lemi import LEMICollection
            >>> lc = LEMICollection("/path/to/single/lemi/station")
            >>> lemi_df = lc.to_dataframe()

        """

        dipole_list = []
        entries = []
        for fn in self.get_files(self.file_ext):
            nims_obj = NIMS(fn)
            nims_obj.read_header()

            entry = {}
            entry["survey"] = self.survey_id
            entry["station"] = nims_obj.station
            entry["run"] = nims_obj.run_id
            entry["start"] = nims_obj.start_time.isoformat()
            entry["end"] = nims_obj.end_time.isoformat()
            entry["channel_id"] = 1
            entry["component"] = ",".join(
                ["hx", "hy", "hz", "ex", "ey", "temperature"]
            )
            entry["fn"] = fn
            entry["sample_rate"] = nims_obj.sample_rate
            entry["file_size"] = nims_obj.file_size
            entry["n_samples"] = nims_obj.n_samples
            entry["sequence_number"] = 0
            entry["instrument_id"] = "NIMS"
            entry["calibration_fn"] = None

            entries.append(entry)

            dipole_list.append(nims_obj.ex_length)
            dipole_list.append(nims_obj.ey_length)

        # make pandas dataframe and set data types
        df = self._sort_df(
            self._set_df_dtypes(pd.DataFrame(entries)), run_name_zeros
        )

        return df

    def assign_run_names(self, df, zeros=2):
        """
        Assign run names assuming a row represents single station

        Run names are assigned as sr{sample_rate}_{run_number:0{zeros}}.

        :param df: Dataframe with the appropriate columns
        :type df: :class:`pandas.DataFrame`
        :param zeros: number of zeros in run name, defaults to 4
        :type zeros: int, optional
        :return: Dataframe with run names
        :rtype: :class:`pandas.DataFrame`

        """

        for station in df.station.unique():
            count = 1
            for row in (
                df[df.station == station].sort_values("start").itertuples()
            ):
                if row.run is None:
                    df.loc[
                        row.Index, "run"
                    ] = f"sr{row.sample_rate}_{count:0{zeros}}"
                df.loc[row.Index, "sequence_number"] = count
                count += 1

        return df



"""
Import all Group objects
"""
## !!! DO NOT CHANGE ORDER !!!
from .base import BaseGroup
from .reports import ReportsGroup
from .standards import StandardsGroup
from .filters import FiltersGroup
from .estimate_dataset import EstimateDataset
from .transfer_function import TransferFunctionGroup
from .master_station_run_channel import (
    MasterStationGroup,
    StationGroup,
    RunGroup,
    ElectricDataset,
    MagneticDataset,
    ChannelDataset,
    AuxiliaryDataset,
)
from .survey import MasterSurveyGroup, SurveyGroup
from .experiment import ExperimentGroup


__all__ = [
    "BaseGroup",
    "ExperimentGroup",
    "SurveyGroup",
    "ReportsGroup",
    "StandardsGroup",
    "StationGroup",
    "MasterStationGroup",
    "MasterSurveyGroup",
    "FiltersGroup",
    "RunGroup",
    "ChannelDataset",
    "ElectricDataset",
    "MagneticDataset",
    "AuxiliaryDataset",
    "EstimateDataset",
    "TransferFunctionGroup",
]



# -*- coding: utf-8 -*-
"""
Created on Wed Jun  9 08:55:16 2021

:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================
from mt_metadata.timeseries.filters import PoleZeroFilter

from mth5.groups.base import BaseGroup

# =============================================================================
# ZPK Group
# =============================================================================
class ZPKGroup(BaseGroup):
    """
    Container for ZPK type filters

    """

    def __init__(self, group, **kwargs):
        super().__init__(group, **kwargs)

    @property
    def filter_dict(self):
        """

        Dictionary of available ZPK filters

        :return: DESCRIPTION
        :rtype: TYPE

        """
        f_dict = {}
        for key in self.hdf5_group.keys():
            zpk_group = self.hdf5_group[key]
            f_dict[key] = {
                "type": zpk_group.attrs["type"],
                "hdf5_ref": zpk_group.ref,
            }

        return f_dict

    def add_filter(self, name, poles, zeros, zpk_metadata):
        """
        create an HDF5 group/dataset from information given.

        :param name: Nane of the filter
        :type name: string
        :param poles: poles of the filter as complex numbers
        :type poles: np.ndarray(dtype=complex)
        :param zeros: zeros of the filter as complex numbers
        :type zeros: np.ndarray(dtype=comples)
        :param zpk_metadata: metadata dictionary see
        :class:`mt_metadata.timeseries.filters.PoleZeroFilter` for details on entries
        :type zpk_metadata: dictionary

        """
        # create a group for the filter by the name
        zpk_filter_group = self.hdf5_group.create_group(name)

        # create datasets for the poles and zeros
        poles_ds = zpk_filter_group.create_dataset(
            "poles",
            poles.shape,
            dtype=complex,
            **self.dataset_options,
        )
        zeros_ds = zpk_filter_group.create_dataset(
            "zeros",
            zeros.shape,
            dtype=complex,
            **self.dataset_options,
        )

        # when filling data need to fill the full row for what ever reason.
        poles_ds[:] = poles
        zeros_ds[:] = zeros

        # fill in the metadata
        zpk_filter_group.attrs.update(zpk_metadata)

        return zpk_filter_group

    def remove_filter(self):
        pass

    def get_filter(self, name):
        """
        Get a filter from the name

        :param name: name of the filter
        :type name: string

        :return: HDF5 group of the ZPK filter
        """
        return self.hdf5_group[name]

    def from_object(self, zpk_object):
        """
        make a filter from a :class:`mt_metadata.timeseries.filters.PoleZeroFilter`

        :param zpk_object: MT metadata PoleZeroFilter
        :type zpk_object: :class:`mt_metadata.timeseries.filters.PoleZeroFilter`

        """

        if not isinstance(zpk_object, PoleZeroFilter):
            msg = f"Filter must be a PoleZeroFilter not {type(zpk_object)}"
            self.logger.error(msg)
            raise TypeError(msg)

        input_dict = zpk_object.to_dict(single=True, required=False)
        input_dict.pop("poles")
        input_dict.pop("zeros")
        for k, v in input_dict.items():
            if v is None:
                input_dict[k] = str(v)

        zpk_group = self.add_filter(
            zpk_object.name,
            zpk_object.poles,
            zpk_object.zeros,
            input_dict,
        )
        return zpk_group

    def to_object(self, name):
        """
        make a :class:`mt_metadata.timeseries.filters.pole_zeros_filter` object

        :return: DESCRIPTION
        :rtype: TYPE

        """

        zpk_group = self.get_filter(name)

        zpk_obj = PoleZeroFilter(**zpk_group.attrs)

        if "poles" in zpk_group.keys():
            if zpk_group["poles"].dtype == complex:
                zpk_obj.poles = zpk_group["poles"][:]
            elif "real" in zpk_group["poles"].dtype.names:
                zpk_obj.poles = (
                    zpk_group["poles"][()]["real"]
                    + 1j * zpk_group["poles"][()]["imag"]
                )
            else:
                raise ValueError(
                    f"Cannot convert values to complex valued poles, check filter {name}"
                )
        else:
            self.logger.debug(f"ZPK filter {name} has no poles")
            zpk_obj.poles = []

        if "zeros" in zpk_group.keys():
            if zpk_group["zeros"].dtype == complex:
                zpk_obj.zeros = zpk_group["zeros"][:]
            elif "real" in zpk_group["zeros"].dtype.names:
                zpk_obj.zeros = (
                    zpk_group["zeros"][()]["real"]
                    + 1j * zpk_group["zeros"][()]["imag"]
                )
            else:
                raise ValueError(
                    f"Cannot convert values to complex valued zeros, check filter {name}"
                )
        else:
            self.logger.debug(f"ZPK filter {name} has no zeros")
            zpk_obj.zeros = []

        return zpk_obj



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 24 11:24:57 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import logging
from mt_metadata.utils.mttime import MTime

# =============================================================================
class Z3DSchedule:
    """
    Will read in the schedule information of a Z3D file and make each metadata
    entry an attirbute. The attributes are left in capitalization of the Z3D file.

    :param fn: full path to Z3D file
    :type fn: string or :class:`pathlib.Path`
    :param fid:  file object ex. open(Z3Dfile, 'rb')
    :type fid: file

    ======================== ==================================================
    Attributes               Definition
    ======================== ==================================================
    AutoGain                 Auto gain for the channel
    Comment                  Any comments for the schedule
    Date                     Date of when the schedule action was started
                             YYYY-MM-DD
    Duty                     Duty cycle of the transmitter
    FFTStacks                FFT stacks from the transmitter
    Filename                 Name of the file that the ZEN gives it
    Gain                     Gain of the channel
    Log                      Log the data [ Y | N ]
    NewFile                  Create a new file [ Y | N ]
    Period                   Period of the transmitter
    RadioOn                  Turn on the radio [ Y | N ]
    SR                       Sampling Rate in Hz
    SamplesPerAcq            Samples per aquisition for transmitter
    Sleep                    Set the box to sleep [ Y | N ]
    Sync                     Sync with GPS [ Y | N ]
    Time                     Time the schedule action started
                             HH:MM:SS (GPS time)
    _header_len              length of header in bits (512)
    _schedule_metadata_len   length of schedule metadata in bits (512)
    fid                      file object of the file
    fn                       file name to read in
    meta_string              string of the schedule
    ======================== ==================================================

    :Example:

        >>> import mtpy.usgs.zen as zen
        >>> Z3Dfn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
        >>> header_obj = zen.Z3DSchedule()
        >>> header_obj.read_schedule()

    """

    def __init__(self, fn=None, fid=None, **kwargs):
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}"
        )
        self.fn = fn
        self.fid = fid
        self.meta_string = None

        self._schedule_metadata_len = 512
        self._header_len = 512

        self.AutoGain = None
        self.Comment = None
        self.Date = None
        self.Duty = None
        self.FFTStacks = None
        self.Filename = None
        self.Gain = None
        self.Log = None
        self.NewFile = None
        self.Period = None
        self.RadioOn = None
        self.SR = None
        self.SamplesPerAcq = None
        self.Sleep = None
        self.Sync = None
        self.Time = None
        self.initial_start = MTime()

        for key in kwargs:
            setattr(self, key, kwargs[key])

    def read_schedule(self, fn=None, fid=None):
        """
        read meta data string
        """
        if fn is not None:
            self.fn = fn

        if fid is not None:
            self.fid = fid

        if self.fn is None and self.fid is None:
            self.logger.warning("No Z3D file to read.")
        elif self.fn is None:
            if self.fid is not None:
                self.fid.seek(self._header_len)
                self.meta_string = self.fid.read(self._header_len)
        elif self.fn is not None:
            if self.fid is None:
                self.fid = open(self.fn, "rb")
                self.fid.seek(self._header_len)
                self.meta_string = self.fid.read(self._header_len)
            else:
                self.fid.seek(self._header_len)
                self.meta_string = self.fid.read(self._header_len)

        meta_list = self.meta_string.split(b"\n")
        for m_str in meta_list:
            m_str = m_str.decode()
            if m_str.find("=") > 0:
                m_list = m_str.split("=")
                m_key = m_list[0].split(".")[1].strip()
                m_key = m_key.replace("/", "")
                m_value = m_list[1].strip()
                setattr(self, m_key, m_value)

        self.initial_start = MTime(
            time=f"{self.Date}T{self.Time}", gps_time=True
        )



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 10 18:19:12 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
from collections import OrderedDict
from pathlib import Path

import pandas as pd
import numpy as np

from mth5.io.nims import NIMSCollection

# =============================================================================


@unittest.skipIf(
    "peacock" not in str(Path(__file__).as_posix()), "local files"
)
class TestNIMSCollection(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.nc = NIMSCollection(r"c:\Users\jpeacock\OneDrive - DOI\mt\nims")

        self.df = self.nc.to_dataframe([8])
        self.runs = self.nc.get_runs([8])

        self.station = self.df.station.unique()[0]

    def test_file_path(self):
        self.assertIsInstance(self.nc.file_path, Path)

    def test_get_files(self):
        self.assertListEqual(
            ["mnp300a.BIN", "mnp300b.BIN"],
            [fn.name for fn in self.nc.get_files(self.nc.file_ext)],
        )

    def test_df_columns(self):
        self.assertListEqual(
            self.nc._columns,
            self.df.columns.to_list(),
        )

    def test_df_shape(self):
        self.assertEqual(self.df.shape, (2, 14))

    def test_df_types(self):
        self.df = self.nc._set_df_dtypes(self.df)
        with self.subTest("start"):
            self.assertTrue(
                self.df.start.dtype.type
                == pd._libs.tslibs.timestamps.Timestamp
            )
        with self.subTest("end"):
            self.assertTrue(
                self.df.end.dtype.type == pd._libs.tslibs.timestamps.Timestamp
            )

        with self.subTest("instrument_id"):
            self.assertTrue(self.df.instrument_id.dtype.type == np.object_)

        with self.subTest("calibration_fn"):
            self.assertTrue(self.df.calibration_fn.dtype.type == np.object_)

    def test_df_run_names(self):
        self.assertListEqual(
            ["mnp300a", "mnp300b"],
            self.df.run.to_list(),
        )

    def test_runs_keys(self):
        self.assertListEqual(
            list(self.runs[self.station].keys()),
            ["mnp300a", "mnp300b"],
        )

    def test_run_dtype(self):
        self.assertIsInstance(self.runs, OrderedDict)

    def test_run_elements(self):
        for key, rdf in self.runs[self.station].items():
            with self.subTest(key):
                self.assertTrue(
                    (self.df[self.df.run == key].eq(rdf).all(axis=0).all())
                )


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 10 18:19:12 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
from collections import OrderedDict
from pathlib import Path

import pandas as pd
import numpy as np

from mth5.io.lemi import LEMICollection

# =============================================================================


@unittest.skipIf(
    "peacock" not in str(Path(__file__).as_posix()), "local files"
)
class TestLEMICollection(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.lc = LEMICollection(
            r"c:\Users\jpeacock\OneDrive - DOI\mt\lemi\DATA0110"
        )
        self.lc.station_id = "mt01"
        self.lc.survey_id = "test"

        self.df = self.lc.to_dataframe()
        self.runs = self.lc.get_runs([1])

    def test_file_path(self):
        self.assertIsInstance(self.lc.file_path, Path)

    def test_get_files(self):
        self.assertListEqual(
            [
                "202009302021.TXT",
                "202009302029.TXT",
                "202009302054.TXT",
                "202009302112.TXT",
                "202009302114.TXT",
                "202010010000.TXT",
                "202010020000.TXT",
                "202010030000.TXT",
                "202010040000.TXT",
                "202010050000.TXT",
                "202010060000.TXT",
                "202010070000.TXT",
            ],
            [fn.name for fn in self.lc.get_files(self.lc.file_ext)],
        )

    def test_df_columns(self):
        self.assertListEqual(
            self.lc._columns,
            self.df.columns.to_list(),
        )

    def test_df_shape(self):
        self.assertEqual(self.df.shape, (12, 14))

    def test_df_types(self):
        self.df = self.lc._set_df_dtypes(self.df)
        with self.subTest("start"):
            self.assertTrue(
                self.df.start.dtype.type
                == pd._libs.tslibs.timestamps.Timestamp
            )
        with self.subTest("end"):
            self.assertTrue(
                self.df.end.dtype.type == pd._libs.tslibs.timestamps.Timestamp
            )

        with self.subTest("instrument_id"):
            self.assertTrue(self.df.instrument_id.dtype.type == np.object_)

        with self.subTest("calibration_fn"):
            self.assertTrue(self.df.calibration_fn.dtype.type == np.object_)

    def test_survey_id(self):
        self.assertTrue((self.df.survey == self.lc.survey_id).all())

    def test_df_run_names(self):
        self.assertListEqual(
            [
                "sr1_0001",
                "sr1_0002",
                "sr1_0003",
                "sr1_0004",
                "sr1_0005",
                "sr1_0005",
                "sr1_0005",
                "sr1_0005",
                "sr1_0005",
                "sr1_0005",
                "sr1_0005",
                "sr1_0005",
            ],
            self.df.run.to_list(),
        )

    def test_runs_keys(self):
        self.assertListEqual(
            list(self.runs[self.lc.station_id].keys()),
            ["sr1_0001", "sr1_0002", "sr1_0003", "sr1_0004", "sr1_0005"],
        )

    def test_run_dtype(self):
        self.assertIsInstance(self.runs, OrderedDict)

    def test_run_elements(self):
        for key, rdf in self.runs[self.lc.station_id].items():
            with self.subTest(key):
                self.assertTrue(
                    (self.df[self.df.run == key].eq(rdf).all(axis=0).all())
                )


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Fri Sep  2 13:50:51 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
from mt_metadata.timeseries.filters import (
    PoleZeroFilter,
    TimeDelayFilter,
    CoefficientFilter,
    ChannelResponseFilter,
)

# =============================================================================


class ResponseError(Exception):
    pass


class Response(object):
    """
    Common NIMS response filters for electric and magnetic channels

    """

    def __init__(self, system_id=None, **kwargs):
        self.system_id = system_id
        self.hardware = "PC"
        self.instrument_type = "backbone"
        self.sample_rate = 1

        self.e_conversion_factor = 409600000.0
        self.h_conversion_factor = 100

        self.time_delays_dict = {
            "hp200": {
                "hx": -0.0055,
                "hy": -0.0145,
                "hz": -0.0235,
                "ex": -0.1525,
                "ey": -0.0275,
            },
            1: {
                "hx": -0.1920,
                "hy": -0.2010,
                "hz": -0.2100,
                "ex": -0.2850,
                "ey": -0.2850,
            },
            8: {
                "hx": -0.2455,
                "hy": -0.2365,
                "hz": -0.2275,
                "ex": -0.1525,
                "ey": -0.1525,
            },
        }

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def magnetic_low_pass(self):
        """
        Low pass 3 pole filter

        :return: DESCRIPTION
        :rtype: TYPE

        """
        return PoleZeroFilter(
            name="nims_3_pole_butterworth",
            zeros=[],
            poles=[
                complex(-6.283185, 10.882477),
                complex(-6.283185, -10.882477),
                complex(-12.566371, 0),
            ],
            units_out="volts",
            units_in="nanotesla",
            normalization_factor=2002.26936395594,
        )

    @property
    def magnetic_conversion(self):
        """

        :return: DESCRIPTION
        :rtype: TYPE

        """

        return CoefficientFilter(
            name="h_analog_to_digital",
            gain=self.h_conversion_factor,
            units_in="volts",
            units_out="digital counts",
        )

    @property
    def electric_low_pass(self):
        """
        5 pole electric low pass filter
        :return: DESCRIPTION
        :rtype: TYPE

        """
        return PoleZeroFilter(
            name="nims_5_pole_butterworth",
            zeros=[],
            poles=[
                complex(-3.88301, 11.9519),
                complex(-3.88301, -11.9519),
                complex(-10.1662, 7.38651),
                complex(-10.1662, -7.38651),
                complex(-12.5664, 0.0),
            ],
            units_in="volts",
            units_out="volts",
            normalization_factor=313383.493219835,
        )

    @property
    def electric_high_pass_pc(self):
        """
        1-pole low pass filter for 8 hz instruments
        :return: DESCRIPTION
        :rtype: TYPE

        """
        return PoleZeroFilter(
            name="nims_1_pole_butterworth",
            zeros=[complex(0.0, 0.0)],
            poles=[complex(-3.333333e-05, 0.0)],
            normalization_factor=1,
            units_in="volts",
            units_out="volts",
        )

    @property
    def electric_high_pass_hp(self):
        """
        1-pole low pass for 1 hz instuments
        :return: DESCRIPTION
        :rtype: TYPE

        """
        return PoleZeroFilter(
            name="nims_1_pole_butterworth",
            zeros=[complex(0.0, 0.0)],
            poles=[complex(-1.66667e-04, 0.0)],
            normalization_factor=1,
            units_in="volts",
            units_out="volts",
        )

    @property
    def electric_conversion(self):
        """
        electric channel conversion from counts to volts
        :return: DESCRIPTION
        :rtype: TYPE

        """
        return CoefficientFilter(
            name="e_analog_to_digital",
            gain=self.e_conversion_factor,
            units_in="volts",
            units_out="digital counts",
        )

    @property
    def electric_physical_units(self):
        """

        :return: DESCRIPTION
        :rtype: TYPE

        """
        return CoefficientFilter(
            name="to_mt_units",
            gain=1e-6,
            units_in="millivolts per kilometer",
            units_out="volts per meter",
        )

    def get_electric_high_pass(self, hardware="pc"):
        """
        get the electric high pass filter based on the hardware
        """

        self.hardware = hardware
        if "pc" in hardware.lower():
            return self.electric_high_pass_pc
        elif "hp" in hardware.lower():
            return self.electric_high_pass_hp
        else:
            raise ResponseError(f"Hardware value {hardware} not understood")

    def _get_dt_filter(self, channel, sample_rate):
        """
        get the DT filter based on channel ans sampling rate
        """
        dt_filter = TimeDelayFilter(
            name=f"{channel}_time_offset",
            delay=self.time_delays_dict[sample_rate][channel],
            units_in="digital counts",
            units_out="digital counts",
        )
        return dt_filter

    def dipole_filter(self, length):
        """
        Make a dipole filter

        :param length: dipole length in meters
        :type length: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        return CoefficientFilter(
            name=f"dipole_{length:.2f}",
            gain=length,
            units_in="volts per meter",
            units_out="volts",
        )

    def _get_magnetic_filter(self, channel):
        """
        get mag filter, seems to be the same no matter what
        """

        return [
            self.magnetic_low_pass,
            self.magnetic_conversion,
            self._get_dt_filter(channel, self.sample_rate),
        ]

    def _get_electric_filter(self, channel, dipole_length):
        """
        Get electric filter
        """

        filter_list = []
        filter_list.append(self.electric_physical_units)
        filter_list.append(self.dipole_filter(dipole_length))
        filter_list.append(self.electric_low_pass)
        if self.instrument_type in ["backbone"]:
            filter_list.append(self.get_electric_high_pass(self.hardware))

        filter_list.append(self.electric_conversion)
        filter_list.append(self._get_dt_filter(channel, self.sample_rate))

        return filter_list

    def get_channel_response(self, channel, dipole_length=1):
        """
        Get the full channel response filter
        :param channel: DESCRIPTION
        :type channel: TYPE
        :param dipole_length: DESCRIPTION, defaults to 1
        :type dipole_length: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if channel[0] in ["e"]:
            return ChannelResponseFilter(
                filters_list=self._get_electric_filter(channel, dipole_length)
            )

        elif channel[0] in ["b", "h"]:
            return ChannelResponseFilter(
                filters_list=self._get_magnetic_filter(channel)
            )

        else:
            raise ValueError(f"Channel {channel} not supported.")



# -*- coding: utf-8 -*-
"""
Tests for MTh5

Created on Thu Jun 18 16:54:19 2020

@author: jpeacock
"""
# =============================================================================
# Imports
# =============================================================================

import unittest
from pathlib import Path

from mth5 import mth5


fn_path = Path(__file__).parent
# =============================================================================
#
# =============================================================================
mth5.helpers.close_open_files()


class TestMTH5Basics(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.mth5_obj = mth5.MTH5()
        self.maxDiff = None

    def test_str(self):
        self.assertEqual(
            self.mth5_obj.__str__(),
            "HDF5 file is closed and cannot be accessed.",
        )

    def test_repr(self):
        self.assertEqual(
            self.mth5_obj.__repr__(),
            "HDF5 file is closed and cannot be accessed.",
        )

    def test_file_type(self):
        self.assertEqual(self.mth5_obj.file_type, "mth5")

    def test_set_file_type_fail(self):
        def set_file_type(value):
            self.mth5_obj.file_type = value

        with self.subTest("bad value"):
            self.assertRaises(ValueError, set_file_type, 10)

        with self.subTest("bad file type"):
            self.assertRaises(ValueError, set_file_type, "asdf")

    def test_set_file_version_fail(self):
        def set_file_version(value):
            self.mth5_obj.file_version = value

        with self.subTest("bad value"):
            self.assertRaises(ValueError, set_file_version, 10)

        with self.subTest("bad file version"):
            self.assertRaises(ValueError, set_file_version, "4")

    def test_set_data_level_fail(self):
        def set_data_level(value):
            self.mth5_obj.data_level = value

        with self.subTest("bad value"):
            self.assertRaises(ValueError, set_data_level, "y")

        with self.subTest("bad data level"):
            self.assertRaises(ValueError, set_data_level, "10")

    def test_filename_fail(self):
        self.mth5_obj.filename = "filename.txt"
        with self.subTest("isinstance path"):
            self.assertIsInstance(self.mth5_obj.filename, Path)

        with self.subTest("extension"):
            self.assertEqual(self.mth5_obj.filename.suffix, ".h5")

    def test_is_read(self):
        self.assertEqual(self.mth5_obj.h5_is_read(), False)

    def test_is_write(self):
        self.assertEqual(self.mth5_obj.h5_is_write(), False)

    def test_validation(self):
        self.assertEqual(self.mth5_obj.validate_file(), False)

    @classmethod
    def tearDownClass(self):
        self.mth5_obj.close_mth5()


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 23 17:03:53 2020

:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================
import numpy as np
import h5py

from mth5.groups.base import BaseGroup

# =============================================================================
# Reports Group
# =============================================================================
class ReportsGroup(BaseGroup):
    """
    Not sure how to handle this yet

    """

    def __init__(self, group, **kwargs):

        super().__init__(group, **kwargs)

        # summary of reports
        self._defaults_summary_attrs = {
            "name": "summary",
            "max_shape": (1000,),
            "dtype": np.dtype(
                [
                    ("name", "S5"),
                    ("type", "S32"),
                    ("summary", "S200"),
                    ("hdf5_reference", h5py.ref_dtype),
                ]
            ),
        }

    def add_report(self, report_name, report_metadata=None, report_data=None):
        """

        :param report_name: DESCRIPTION
        :type report_name: TYPE
        :param report_metadata: DESCRIPTION, defaults to None
        :type report_metadata: TYPE, optional
        :param report_data: DESCRIPTION, defaults to None
        :type report_data: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """
        self.logger.error("Not Implemented yet")



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 23 17:05:33 2020

:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================
import inspect
import numpy as np

from mth5.groups.base import BaseGroup
from mth5.tables import MTH5Table
from mth5.utils.exceptions import MTH5TableError

from mt_metadata.base import BaseDict
from mt_metadata import timeseries
from mt_metadata.timeseries import filters
from mt_metadata.utils.validators import validate_attribute

ts_classes = dict(inspect.getmembers(timeseries, inspect.isclass))
flt_classes = dict(inspect.getmembers(filters, inspect.isclass))
# =============================================================================
# Summarize standards
# =============================================================================
def summarize_metadata_standards():
    """
    Summarize metadata standards into a dictionary
    """
    # need to be sure to make copies otherwise things will get
    # added in not great places.
    # need to be sure to make copies otherwise things will get
    # added in not great places.
    summary_dict = BaseDict()
    for key in ["survey", "station", "run", "electric", "magnetic", "auxiliary"]:
        obj = ts_classes[key.capitalize()]()
        summary_dict.add_dict(obj._attr_dict.copy(), key)

    for key in [
        "Coefficient",
        "FIR",
        "FrequencyResponseTable",
        "PoleZero",
        "TimeDelay",
    ]:
        key += "Filter"
        obj = flt_classes[key]()
        summary_dict.add_dict(obj._attr_dict.copy(), validate_attribute(key))
    return summary_dict


# =============================================================================
# Standards Group
# =============================================================================


class StandardsGroup(BaseGroup):
    """
    The StandardsGroup is a convenience group that stores the metadata
    standards that were used to make the current file.  This is to help a
    user understand the metadata directly from the file and not have to look
    up documentation that might not be updated.

    The metadata standards are stored in the summary table
    ``/Survey/Standards/summary``

    >>> standards = mth5_obj.standards_group
    >>> standards.summary_table
    index | attribute | type | required | style | units | description |
    options  |  alias |  example
    --------------------------------------------------------------------------

    """

    def __init__(self, group, **kwargs):

        super().__init__(group, **kwargs)

        self._defaults_summary_attrs = {
            "name": "summary",
            "max_shape": (500,),
            "dtype": np.dtype(
                [
                    ("attribute", "S72"),
                    ("type", "S15"),
                    ("required", np.bool_),
                    ("style", "S72"),
                    ("units", "S32"),
                    ("description", "S300"),
                    ("options", "S150"),
                    ("alias", "S72"),
                    ("example", "S72"),
                    ("default", "S72"),
                ]
            ),
        }

    @property
    def summary_table(self):
        return MTH5Table(self.hdf5_group["summary"])

    def get_attribute_information(self, attribute_name):
        """
        get information about an attribute

        The attribute name should be in the summary table.

        :param attribute_name: attribute name
        :type attribute_name: string
        :return: prints a description of the attribute
        :raises MTH5TableError:  if attribute is not found

        >>> standars = mth5_obj.standards_group
        >>> standards.get_attribute_information('survey.release_license')
        survey.release_license
        --------------------------
                type          : string
                required      : True
                style         : controlled vocabulary
                units         :
                description   : How the data can be used. The options are based on
                         Creative Commons licenses. For details visit
                         https://creativecommons.org/licenses/
                options       : CC-0,CC-BY,CC-BY-SA,CC-BY-ND,CC-BY-NC-SA,CC-BY-NC-ND
                alias         :
                example       : CC-0
                default       : CC-0

        """
        find = self.summary_table.locate("attribute", attribute_name)
        if len(find) == 0:
            msg = f"Could not find {attribute_name} in standards."
            self.logger.error(msg)
            raise MTH5TableError(msg)

        meta_item = self.summary_table.array[find]
        lines = ["", attribute_name, "-" * (len(attribute_name) + 4)]
        for name, value in zip(meta_item.dtype.names[1:], meta_item.item()[1:]):
            if isinstance(value, (bytes, np.bytes_)):
                value = value.decode()
            lines.append("\t{0:<14} {1}".format(name + ":", value))

        print("\n".join(lines))

    def summary_table_from_dict(self, summary_dict):
        """
        Fill summary table from a dictionary that summarizes the metadata
        for the entire survey.

        :param summary_dict: Flattened dictionary of all metadata standards
                             within the survey.
        :type summary_dict: dictionary

        """

        for key, v_dict in summary_dict.items():
            key_list = [key]
            for dkey in self.summary_table.dtype.names[1:]:
                value = v_dict[dkey]

                if isinstance(value, list):
                    if len(value) == 0:
                        value = ""

                    else:
                        value = ",".join(["{0}".format(ii) for ii in value])
                if value is None:
                    value = ""

                key_list.append(value)

            key_list = np.array([tuple(key_list)], self.summary_table.dtype)
            index = self.summary_table.add_row(key_list)

        self.logger.debug(f"Added {index} rows to Standards Group")

    def initialize_group(self):
        """
        Initialize the group by making a summary table that summarizes
        the metadata standards used to describe the data.

        Also, write generic metadata information.

        """
        if self.dataset_options["compression"] is None:
            summary_dataset = self.hdf5_group.create_dataset(
                self._defaults_summary_attrs["name"],
                (0,),
                maxshape=self._defaults_summary_attrs["max_shape"],
                dtype=self._defaults_summary_attrs["dtype"],
            )
        else:
            summary_dataset = self.hdf5_group.create_dataset(
                self._defaults_summary_attrs["name"],
                (0,),
                maxshape=self._defaults_summary_attrs["max_shape"],
                dtype=self._defaults_summary_attrs["dtype"],
                **self.dataset_options,
            )

        summary_dataset.attrs.update(
            {
                "type": "summary table",
                "last_updated": "date_time",
                "reference": summary_dataset.ref,
            }
        )

        self.logger.debug(
            "Created %s table with max_shape = %s, dtype=%s",
            self._defaults_summary_attrs["name"],
            self._defaults_summary_attrs["max_shape"],
            self._defaults_summary_attrs["dtype"],
        )
        self.logger.debug(
            "used options: "
            + "; ".join(["%s = %s" % (k, v) for k, v in self.dataset_options.items()])
        )

        self.summary_table_from_dict(summarize_metadata_standards())

        self.write_metadata()



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 12 12:30:32 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
import logging

from mth5.io import reader
from mth5.utils.mth5_logger import setup_logger

# =============================================================================
logger = setup_logger("test_reader")


class TestReader(unittest.TestCase):
    def test_get_reader_bin(self):
        key, reader_function = reader.get_reader("bin")
        self.assertEqual(key, "nims")

    def test_get_reader_z3d(self):
        key, reader_function = reader.get_reader("z3d")
        self.assertEqual(key, "zen")

    def test_get_reader_asc(self):
        key, reader_function = reader.get_reader("asc")
        self.assertEqual(key, "usgs_ascii")

    def test_get_reader_txt(self):
        key, reader_function = reader.get_reader("txt")
        self.assertEqual(key, "lemi424")

    def test_get_reader_td_150(self):
        key, reader_function = reader.get_reader("td_150")
        self.assertEqual(key, "phoenix")

    def test_get_reader_td_24k(self):
        key, reader_function = reader.get_reader("td_24k")
        self.assertEqual(key, "phoenix")

    def test_get_reader_miniseed(self):
        key, reader_function = reader.get_reader("mseed")
        self.assertEqual(key, "miniseed")

    def test_get_reader_fail(self):
        self.assertRaises(ValueError, reader.get_reader, "dat")

    def test_input_fail(self):
        self.assertRaises(IOError, reader.read_file, "y.txt")

    def test_input_fail_list(self):
        self.assertRaises(IOError, reader.read_file, ["y.txt"])


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
===============
NIMS
===============

    * deals with reading in NIMS DATA.BIN files
    
    This is a translation from Matlab codes written and edited by:
        * Anna Kelbert
        * Paul Bedrosian
        * Esteban Bowles-Martinez
        * Possibly others.
        
    I've tested it against a version, and it matches.  The data/GPS  gaps I
    still don't understand so for now the time series is just 
    made continuous and the number of missing seconds is clipped from the 
    end of the time series.
    
    .. note:: this only works for 8Hz data for now
    
    
:copyright:
    Jared Peacock (jpeacock@usgs.gov)
    
:license: 
    MIT
"""

# =============================================================================
# Imports
# =============================================================================
import struct
import datetime

import numpy as np
import pandas as pd

from mth5.io.nims.gps import GPS
from mth5.io.nims.header import NIMSHeader
from mth5.io.nims.response_filters import Response
from mth5 import timeseries

from mt_metadata.utils.mttime import MTime
from mt_metadata.timeseries import Station, Run, Electric, Magnetic, Auxiliary

# =============================================================================
# Exceptions
# =============================================================================


class NIMS(NIMSHeader):
    """
    NIMS Class will read in a NIMS DATA.BIN file.

    A fast way to read the binary files are to first read in the GPS strings,
    the third byte in each block as a character and parse that into valid
    GPS stamps.

    Then read in the entire data set as unsigned 8 bit integers and reshape
    the data to be n seconds x block size.  Then parse that array into the
    status information and data.

    I only have a limited amount of .BIN files to test so this will likely
    break if there are issues such as data gaps.  This has been tested against the
    matlab program loadNIMS by Anna Kelbert and the match for all the .bin files
    I have.  If something looks weird check it against that program.

    .. todo:: deal with timing issues, right now a warning is sent to the user
              need to figure out a way to find where the gap is and adjust
              accordingly.

    .. warning::
        Currently Only 8 Hz data is supported

    """

    def __init__(self, fn=None):
        super().__init__(fn)

        # change thes if the sample rate is different
        self.block_size = 131
        self.block_sequence = [1, self.block_size]
        self.sample_rate = 8  ### samples/second
        self.e_conversion_factor = 2.44141221047903e-06
        self.h_conversion_factor = 0.01
        self.t_conversion_factor = 70
        self.t_offset = 18048
        self._int_max = 8388608
        self._int_factor = 16777216
        self._block_dict = {
            "soh": 0,
            "block_len": 1,
            "status": 2,
            "gps": 3,
            "sequence": 4,
            "box_temp": (5, 6),
            "head_temp": (7, 8),
            "logic": 81,
            "end": 130,
        }
        self.info_array = None
        self.stamps = None
        self.ts_data = None
        self.gaps = None
        self.duplicate_list = None

        self._raw_string = None

        self.indices = self._make_index_values()

    def __str__(self):
        lines = [f"NIMS Station: {self.site_name}", "-" * 20]
        lines.append(f"NIMS ID:         {self.box_id}")
        lines.append(f"magnetometer ID: {self.mag_id}")
        lines.append(f"operator:        {self.operator}")
        lines.append(f"location:        {self.state_province}, {self.country}")
        lines.append(f"latitude:        {self.latitude} (degrees)")
        lines.append(f"longitude:       {self.longitude} (degrees)")
        lines.append(f"elevation:       {self.elevation} m")
        lines.append(f"gps stamp:       {self.header_gps_stamp}")
        lines.append(
            f"EX: length = {self.ex_length} m; azimuth = {self.ex_azimuth}"
        )
        lines.append(
            f"EY: length = {self.ey_length} m; azimuth = {self.ey_azimuth}"
        )
        lines.append(f"comments:        {self.comments}")

        if self.has_data():
            lines.append("")
            lines.append(f"Start:      {self.start_time.isoformat()}")
            lines.append(f"End:        {self.end_time.isoformat()}")
            lines.append(f"Data shape: {self.ts_data.shape}")
            lines.append(f"Found {len(self.stamps)} GPS stamps")

        return "\n".join(lines)

    def __repr__(self):
        return self.__str__()

    def has_data(self):
        if self.ts_data is not None:
            return True
        return False

    @property
    def n_samples(self):
        if self.has_data():
            return self.ts_data.shape[0]
        elif self.fn is not None:
            return int(self.file_size / 16.375)

    @property
    def latitude(self):
        """
        median latitude value from all the GPS stamps in decimal degrees
        WGS84

        Only get from the GPRMC stamp as they should be duplicates
        """
        if self.stamps is not None:
            latitude = np.zeros(len(self.stamps))
            for ii, stamp in enumerate(self.stamps):
                latitude[ii] = stamp[1][0].latitude
            return np.median(latitude[np.nonzero(latitude)])
        return self.header_gps_latitude

    @property
    def longitude(self):
        """
        median longitude value from all the GPS stamps in decimal degrees
        WGS84

        Only get from the first stamp within the sets
        """
        if self.stamps is not None:
            longitude = np.zeros(len(self.stamps))
            for ii, stamp in enumerate(self.stamps):
                longitude[ii] = stamp[1][0].longitude
            return np.median(longitude[np.nonzero(longitude)])
        return self.header_gps_longitude

    @property
    def elevation(self):
        """
        median elevation value from all the GPS stamps in decimal degrees
        WGS84

        Only get from the first stamp within the sets
        """
        if self.stamps is not None:
            elevation = np.zeros(len(self.stamps))
            for ii, stamp in enumerate(self.stamps):
                if len(stamp[1]) == 1:
                    elev = stamp[1][0].elevation
                if len(stamp[1]) == 2:
                    elev = stamp[1][1].elevation
                if elev is None:
                    continue
                elevation[ii] = elev
            return np.median(elevation[np.nonzero(elevation)])
        return self.header_gps_elevation

    @property
    def declination(self):
        """
        median elevation value from all the GPS stamps in decimal degrees
        WGS84

        Only get from the first stamp within the sets
        """
        if self.stamps is not None:
            declination = np.zeros(len(self.stamps))
            for ii, stamp in enumerate(self.stamps):
                if stamp[1][0].gps_type == "GPRMC":
                    dec = stamp[1][0].declination
                if dec is None:
                    continue
                declination[ii] = dec
            return np.median(declination[np.nonzero(declination)])
        return None

    @property
    def start_time(self):
        """
        start time is the first good GPS time stamp minus the seconds to the
        beginning of the time series.
        """
        if self.stamps is not None:
            return MTime(self.ts_data.index[0])
        return self.header_gps_stamp

    @property
    def end_time(self):
        """
        start time is the first good GPS time stamp minus the seconds to the
        beginning of the time series.
        """
        if self.stamps is not None:
            return MTime(self.ts_data.index[-1])
        self.logger.warning("Estimating end time from n_samples")
        return self.start_time + int(self.n_samples / self.sample_rate)

    @property
    def box_temperature(self):
        """data logger temperature, sampled at 1 second"""

        if self.ts_data is not None:
            auxiliary_metadata = Auxiliary()
            auxiliary_metadata.from_dict(
                {
                    "channel_number": 6,
                    "component": "temperature",
                    "measurement_azimuth": 0,
                    "measurement_tilt": 0,
                    "sample_rate": 1,
                    "time_period.start": self.start_time.isoformat(),
                    "time_period.end": self.end_time.isoformat(),
                    "type": "auxiliary",
                    "units": "celsius",
                }
            )

            temp = timeseries.ChannelTS(
                channel_type="auxiliary",
                data=self.info_array["box_temp"],
                channel_metadata=auxiliary_metadata,
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
            )
            # interpolate temperature onto the same sample rate as the channels.
            temp._ts = temp._ts.interp_like(self.hx._ts)
            temp.channel_metadata.sample_rate = self.sample_rate
            temp.channel_metadata.time_period.end = self.end_time.isoformat()

            return temp
        return None

    def get_channel_response(self, channel, dipole_length=1):
        """
        Get the channel response for a given channel

        :param channel: DESCRIPTION
        :type channel: TYPE
        :param dipole_length: DESCRIPTION, defaults to 1
        :type dipole_length: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        nims_filters = Response(sample_rate=self.sample_rate)
        return nims_filters.get_channel_response(
            channel, dipole_length=dipole_length
        )

    @property
    def hx_metadata(self):
        if self.ts_data is not None:
            hx_metadata = Magnetic()
            hx_metadata.from_dict(
                {
                    "channel_number": 1,
                    "component": "hx",
                    "measurement_azimuth": 0,
                    "measurement_tilt": 0,
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start_time.isoformat(),
                    "time_period.end": self.end_time.isoformat(),
                    "type": "magnetic",
                    "units": "counts",
                    "sensor.id": self.mag_id,
                    "sensor.manufacturer": "Barry Narod",
                    "sensor.type": "fluxgate triaxial magnetometer",
                }
            )
            return hx_metadata

    @property
    def hx(self):
        """HX"""
        if self.ts_data is not None:

            return timeseries.ChannelTS(
                channel_type="magnetic",
                data=self.ts_data.hx.to_numpy(),
                channel_metadata=self.hx_metadata,
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
                channel_response_filter=self.get_channel_response("hx"),
            )
        return None

    @property
    def hy_metadata(self):
        if self.ts_data is not None:
            hy_metadata = Magnetic()
            hy_metadata.from_dict(
                {
                    "channel_number": 2,
                    "component": "hy",
                    "measurement_azimuth": 90,
                    "measurement_tilt": 0,
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start_time.isoformat(),
                    "time_period.end": self.end_time.isoformat(),
                    "type": "magnetic",
                    "units": "counts",
                    "sensor.id": self.mag_id,
                    "sensor.manufacturer": "Barry Narod",
                    "sensor.type": "fluxgate triaxial magnetometer",
                }
            )
            return hy_metadata

    @property
    def hy(self):
        """HY"""
        if self.ts_data is not None:
            return timeseries.ChannelTS(
                channel_type="magnetic",
                data=self.ts_data.hy.to_numpy(),
                channel_metadata=self.hy_metadata,
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
                channel_response_filter=self.get_channel_response("hy"),
            )
        return None

    @property
    def hz_metadata(self):
        if self.ts_data is not None:
            hz_metadata = Magnetic()
            hz_metadata.from_dict(
                {
                    "channel_number": 3,
                    "component": "hz",
                    "measurement_azimuth": 0,
                    "measurement_tilt": 0,
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start_time.isoformat(),
                    "time_period.end": self.end_time.isoformat(),
                    "type": "magnetic",
                    "units": "counts",
                    "sensor.id": self.mag_id,
                    "sensor.manufacturer": "Barry Narod",
                    "sensor.type": "fluxgate triaxial magnetometer",
                }
            )
            return hz_metadata

    @property
    def hz(self):
        """HZ"""
        if self.ts_data is not None:

            return timeseries.ChannelTS(
                channel_type="magnetic",
                data=self.ts_data.hz.to_numpy(),
                channel_metadata=self.hz_metadata,
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
                channel_response_filter=self.get_channel_response("hz"),
            )
        return None

    @property
    def ex_metadata(self):
        if self.ts_data is not None:
            ex_metadata = Electric()
            ex_metadata.from_dict(
                {
                    "channel_number": 4,
                    "component": "ex",
                    "measurement_azimuth": self.ex_azimuth,
                    "measurement_tilt": 0,
                    "sample_rate": self.sample_rate,
                    "dipole_length": self.ex_length,
                    "time_period.start": self.start_time.isoformat(),
                    "time_period.end": self.end_time.isoformat(),
                    "type": "electric",
                    "units": "counts",
                    "negative.id": self.s_electrode_id,
                    "positive.id": self.n_electrode_id,
                }
            )

            return ex_metadata

    @property
    def ex(self):
        """EX"""
        if self.ts_data is not None:

            return timeseries.ChannelTS(
                channel_type="electric",
                data=self.ts_data.ex.to_numpy(),
                channel_metadata=self.ex_metadata,
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
                channel_response_filter=self.get_channel_response(
                    "ex", self.ex_length
                ),
            )
        return None

    @property
    def ey_metadata(self):
        if self.ts_data is not None:
            ey_metadata = Electric()
            ey_metadata.from_dict(
                {
                    "channel_number": 5,
                    "component": "ey",
                    "measurement_azimuth": self.ey_azimuth,
                    "measurement_tilt": 0,
                    "sample_rate": self.sample_rate,
                    "dipole_length": self.ey_length,
                    "time_period.start": self.start_time.isoformat(),
                    "time_period.end": self.end_time.isoformat(),
                    "type": "electric",
                    "units": "counts",
                    "negative.id": self.w_electrode_id,
                    "positive.id": self.e_electrode_id,
                }
            )

            return ey_metadata

    @property
    def ey(self):
        """EY"""
        if self.ts_data is not None:

            return timeseries.ChannelTS(
                channel_type="electric",
                data=self.ts_data.ey.to_numpy(),
                channel_metadata=self.ey_metadata,
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
                channel_response_filter=self.get_channel_response(
                    "ey", self.ey_length
                ),
            )

        return None

    @property
    def run_metadata(self):
        """Run metadata"""

        if self.ts_data is not None:
            run_metadata = Run()
            run_metadata.from_dict(
                {
                    "Run": {
                        "comments": self.comments,
                        "data_logger.firmware.author": "B. Narod",
                        "data_logger.firmware.name": "nims",
                        "data_logger.firmware.version": "1.0",
                        "data_logger.manufacturer": "Narod",
                        "data_logger.model": self.box_id,
                        "data_logger.id": self.box_id,
                        "data_logger.type": "long period",
                        "id": self.run_id,
                        "data_type": "MTLP",
                        "sample_rate": self.sample_rate,
                        "time_period.end": self.end_time.isoformat(),
                        "time_period.start": self.start_time.isoformat(),
                    }
                }
            )
            for comp in ["hx", "hy", "hz", "ex", "ey"]:
                run_metadata.channels.append(getattr(self, f"{comp}_metadata"))

            return run_metadata

        return None

    @property
    def station_metadata(self):
        """Station metadata from nims file"""
        if self.ts_data is not None:
            station_metadata = Station()
            station_metadata.from_dict(
                {
                    "Station": {
                        "geographic_name": f"{self.site_name}, {self.state_province}, {self.country}",
                        "location.declination.value": self.declination,
                        "location.elevation": self.elevation,
                        "location.latitude": self.latitude,
                        "location.longitude": self.longitude,
                        "id": self.run_id[0:-1],
                        "orientation.reference_frame": "geomagnetic",
                    }
                }
            )
            station_metadata.runs.append(self.run_metadata)
            return station_metadata
        return None

    def to_runts(self, calibrate=False):
        """Get xarray for run"""

        if self.ts_data is not None:
            run = timeseries.RunTS(
                array_list=[
                    self.hx,
                    self.hy,
                    self.hz,
                    self.ex,
                    self.ey,
                    self.box_temperature,
                ],
                run_metadata=self.run_metadata,
                station_metadata=self.station_metadata,
            )
            if calibrate:
                return run.calibrate()

            else:
                return run

        return None

    def _make_index_values(self):
        """
        Index values for the channels recorded
        """
        ### make an array of index values for magnetics and electrics
        indices = np.zeros((8, 5), dtype=int)
        for kk in range(8):
            ### magnetic blocks
            for ii in range(3):
                indices[kk, ii] = 9 + (kk) * 9 + (ii) * 3
            ### electric blocks
            for ii in range(2):
                indices[kk, 3 + ii] = 82 + (kk) * 6 + (ii) * 3
        return indices

    def _get_gps_string_list(self, nims_string):
        """
        get the gps strings from the raw string output by the NIMS.  This will
        take the 3rd value in each block, concatenate into a long string and
        then make a list by splitting by '$'.  The index values of where the
        '$' are found are also calculated.

        :param str nims_string: raw binary string output by NIMS

        :returns: list of index values associated with the location of the '$'

        :returns: list of possible raw GPS strings

        .. note:: This assumes that there are an even amount of data blocks.
                  Might be a bad assumption
        """
        ### get index values of $ and gps_strings
        index_values = []
        gps_str_list = []
        for ii in range(int(len(nims_string) / self.block_size)):
            index = ii * self.block_size + 3
            g_char = struct.unpack("c", nims_string[index : index + 1])[0]
            if g_char == b"$":
                index_values.append((index - 3) / self.block_size)
            gps_str_list.append(g_char)
        gps_raw_stamp_list = b"".join(gps_str_list).split(b"$")
        return index_values, gps_raw_stamp_list

    def get_stamps(self, nims_string):
        """
        get a list of valid GPS strings and match synchronous GPRMC with GPGGA
        stamps if possible.

        :param str nims_string: raw GPS string output by NIMS
        """
        ### read in GPS strings into a list to be parsed later
        index_list, gps_raw_stamp_list = self._get_gps_string_list(nims_string)

        gprmc_list = []
        gpgga_list = []
        ### note we are skipping the first entry, it tends to be not
        ### complete anyway
        for ii, index, raw_stamp in zip(
            range(len(index_list)), index_list, gps_raw_stamp_list[1:]
        ):
            gps_obj = GPS(raw_stamp, index)
            if gps_obj.valid:
                if gps_obj.gps_type == "GPRMC":
                    gprmc_list.append(gps_obj)
                elif gps_obj.gps_type == "GPGGA":
                    gpgga_list.append(gps_obj)
            else:
                self.logger.debug(
                    f"GPS Error: file index {index}, stamp number {ii}"
                )
                max_len = min([len(raw_stamp), 15])
                self.logger.debug(f"GPS Raw Stamp: {raw_stamp[0:max_len]}")

        return self._gps_match_gprmc_gpgga_strings(gprmc_list, gpgga_list)

    def _gps_match_gprmc_gpgga_strings(self, gprmc_list, gpgga_list):
        """
        match GPRMC and GPGGA strings together into a list

        [[GPRMC, GPGGA], ...]

        :param list gprmc_list: list of GPS objects for the GPRMC stamps
        :param list gpgga_list: list of GPS objects for the GPGGA stamps

        :returns: list of matched GPRMC and GPGGA stamps

        """
        ### match up the GPRMC and GPGGA together
        gps_match_list = []
        for gprmc in gprmc_list:
            find = False
            for ii, gpgga in enumerate(gpgga_list):
                if gprmc.time_stamp.time() == gpgga.time_stamp.time():
                    gps_match_list.append([gprmc, gpgga])
                    find = True
                    del gpgga_list[ii]
                    break
            if not find:
                gps_match_list.append([gprmc])

        return gps_match_list

    def _get_gps_stamp_indices_from_status(self, status_array):
        """
        get the index location of the stamps from the status array assuming
        that 0 indicates GPS lock.

        :param :class:`np.ndarray` status_array: an array of status values from data blocks

        :returns: array of index values where GPS lock was acquired ignoring
                  sequential locks.
        """

        index_values = np.where(status_array == 0)[0]
        status_index = np.zeros_like(index_values)
        for ii in range(index_values.size):
            if index_values[ii] - index_values[ii - 1] == 1:
                continue
            else:
                status_index[ii] = index_values[ii]
        status_index = status_index[np.nonzero(status_index)]

        return status_index

    def match_status_with_gps_stamps(self, status_array, gps_list):
        """
        Match the index values from the status array with the index values of
        the GPS stamps.  There appears to be a bit of wiggle room between when the
        lock is recorded and the stamp was actually recorded.  This is typically 1
        second and sometimes 2.

        :param array status_array: array of status values from each data block
        :param list gps_list: list of valid GPS stamps [[GPRMC, GPGGA], ...]

        .. note:: I think there is a 2 second gap between the lock and the
                  first stamp character.
        """

        stamp_indices = self._get_gps_stamp_indices_from_status(status_array)
        gps_stamps = []
        for index in stamp_indices:
            stamp_find = False
            for ii, stamps in enumerate(gps_list):
                index_diff = stamps[0].index - index
                ### check the index value, should be 2 or 74, if it is off by
                ### a value left or right apply a correction.
                if index_diff == 1 or index_diff == 73:
                    index += 1
                    stamps[0].index += 1
                elif index_diff == 2 or index_diff == 74:
                    index = index
                elif index_diff == 3 or index_diff == 75:
                    index -= 1
                    stamps[0].index -= 1
                elif index_diff == 4 or index_diff == 76:
                    index -= 2
                    stamps[0].index -= 2
                if stamps[0].gps_type in ["GPRMC", "gprmc"]:
                    if index_diff in [1, 2, 3, 4]:
                        gps_stamps.append((index, stamps))
                        stamp_find = True
                        del gps_list[ii]
                        break
                elif stamps[0].gps_type in ["GPGGA", "gpgga"]:
                    if index_diff in [73, 74, 75, 76]:
                        gps_stamps.append((index, stamps))
                        stamp_find = True
                        del gps_list[ii]
                        break
            if not stamp_find:
                self.logger.debug(
                    f"GPS Error: No good GPS stamp at {index} seconds"
                )

        return gps_stamps

    def find_sequence(self, data_array, block_sequence=None):
        """
        find a sequence in a given array

        :param array data_array: array of the data with shape [n, m]
                                 where n is the number of seconds recorded
                                 m is the block length for a given sampling
                                 rate.
        :param list block_sequence: sequence pattern to locate
                                    *default* is [1, 131] the start of a
                                    data block.

        :returns: array of index locations where the sequence is found.
        """
        if block_sequence is not None:
            self.block_sequence = block_sequence

        # want to find the index there the test data is equal to the test sequence
        t = np.vstack(
            [
                np.roll(data_array, shift)
                for shift in -np.arange(len(self.block_sequence))
            ]
        ).T
        return np.where(np.all(t == self.block_sequence, axis=1))[0]

    def unwrap_sequence(self, sequence):
        """
        unwrap the sequence to be sequential numbers instead of modulated by
        256.  sets the first number to 0

        :param list sequence: sequence of bytes numbers
        :return: unwrapped number of counts

        """
        count = 0
        unwrapped = np.zeros_like(sequence)
        for ii, seq in enumerate(sequence):
            unwrapped[ii] = seq + count * 256
            if seq == 255:
                count += 1

        unwrapped -= unwrapped[0]

        return unwrapped

    def _locate_duplicate_blocks(self, sequence):
        """
        locate the sequence number where the duplicates exist

        :param list sequence: sequence to match duplicate numbers.
        :returns: list of duplicate index values.
        """

        duplicates = np.where(np.abs(np.diff(sequence)) == 0)[0]
        if len(duplicates) == 0:
            return None
        duplicate_list = []
        for dup in duplicates:
            dup_dict = {}
            dup_dict["sequence_index"] = dup
            dup_dict["ts_index_0"] = dup * self.sample_rate
            dup_dict["ts_index_1"] = dup * self.sample_rate + self.sample_rate
            dup_dict["ts_index_2"] = (dup + 1) * self.sample_rate
            dup_dict["ts_index_3"] = (
                dup + 1
            ) * self.sample_rate + self.sample_rate
            duplicate_list.append(dup_dict)
        return duplicate_list

    def _check_duplicate_blocks(self, block_01, block_02, info_01, info_02):
        """
        make sure the blocks are truly duplicates

        :param np.array block_01: block of data to compare
        :param np.array block_02: block of data to compare
        :param np.array info_01: information array from info_array[sequence_index]
        :param np.array info_02: information array from info_array[sequence_index]

        :returns: boolean if the blocks and information match

        """
        if np.array_equal(block_01, block_02):
            if np.array_equal(info_01, info_02):
                return True
            else:
                return False
        else:
            return False

    def remove_duplicates(self, info_array, data_array):
        """
        remove duplicate blocks, removing the first duplicate as suggested by
        Paul and Anna. Checks to make sure that the mag data are identical for
        the duplicate blocks.  Removes the blocks from the information and
        data arrays and returns the reduced arrays.  This should sync up the
        timing of GPS stamps and index values.

        :param np.array info_array: structured array of block information
        :param np.array data_array: structured array of the data

        :returns: reduced information array
        :returns: reduced data array
        :returns: index of duplicates in raw data

        """
        ### locate
        duplicate_test_list = self._locate_duplicate_blocks(
            self.info_array["sequence"]
        )
        if duplicate_test_list is None:
            return info_array, data_array, None

        duplicate_list = []
        for d in duplicate_test_list:
            if self._check_duplicate_blocks(
                data_array[d["ts_index_0"] : d["ts_index_1"]],
                data_array[d["ts_index_2"] : d["ts_index_3"]],
                info_array[d["sequence_index"]],
                info_array[d["sequence_index"] + 1],
            ):
                duplicate_list.append(d)

        self.logger.debug(f"Deleting {len(duplicate_list)} duplicate blocks")
        ### get the index of the blocks to be removed, namely the 1st duplicate
        ### block
        remove_sequence_index = [d["sequence_index"] for d in duplicate_list]
        remove_data_index = np.array(
            [
                np.arange(d["ts_index_0"], d["ts_index_1"], 1)
                for d in duplicate_list
            ]
        ).flatten()
        ### remove the data
        return_info_array = np.delete(info_array, remove_sequence_index)
        return_data_array = np.delete(data_array, remove_data_index)

        ### set sequence to be monotonic
        return_info_array["sequence"][:] = np.arange(return_info_array.shape[0])

        return return_info_array, return_data_array, duplicate_list

    def read_nims(self, fn=None):
        """
        Read NIMS DATA.BIN file.

        1. Read in the header information and stores those as attributes
           with the same names as in the header file.

        2. Locate the beginning of the data blocks by looking for the
           first [1, 131, ...] combo.  Anything before that is cut out.

        3. Make sure the data is a multiple of the block length, if the
           data is longer the extra bits are cut off.

        4. Read in the GPS data (3rd byte of each block) as characters.
           Parses those into valid GPS stamps with appropriate index locations
           of where the '$' was found.

        5. Read in the data as unsigned 8-bit integers and reshape the array
           into [N, data_block_length].  Parse this array into the status
           information and the data.

        6. Remove duplicate blocks, by removing the first of the duplicates
           as suggested by Anna and Paul.

        7. Match the GPS locks from the status with valid GPS stamps.

        8. Check to make sure that there is the correct number of seconds
           between the first and last GPS stamp.  The extra seconds are cut
           off from the end of the time series.  Not sure if this is the
           best way to accommodate gaps in the data.

        .. note:: The data and information array returned have the duplicates
                  removed and the sequence reset to be monotonic.

        :param str fn: full path to DATA.BIN file

        :Example:

        >>> from mth5.io import nims
        >>> n = nims.NIMS(r"/home/mt_data/nims/mt001.bin")


        """

        if fn is not None:
            self.fn = fn

        st = datetime.datetime.now()
        ### read in header information and get the location of end of header
        self.read_header(self.fn)

        ### load in the entire file, its not too big, start from the
        ### end of the header information.
        with open(self.fn, "rb") as fid:
            fid.seek(self.data_start_seek)
            self._raw_string = fid.read()

        ### read in full string as unsigned integers
        data = np.frombuffer(self._raw_string, dtype=np.uint8)

        ### need to make sure that the data starts with a full block
        find_first = self.find_sequence(data[0 : self.block_size * 5])[0]
        data = data[find_first:]

        ### get GPS stamps from the binary string first
        self.gps_list = self.get_stamps(self._raw_string[find_first:])

        ### check the size of the data, should have an equal amount of blocks
        if (data.size % self.block_size) != 0:
            self.logger.warning(
                f"odd number of bytes {data.size}, not even blocks "
                + "cutting down the data by {0} bits".format(
                    data.size % self.block_size
                )
            )
            end_data = data.size - (data.size % self.block_size)
            data = data[0:end_data]

        # resized the data into an even amount of blocks
        data = data.reshape((int(data.size / self.block_size), self.block_size))

        ### need to parse the data
        ### first get the status information
        self.info_array = np.zeros(
            data.shape[0],
            dtype=[
                ("soh", int),
                ("block_len", int),
                ("status", int),
                ("gps", int),
                ("sequence", int),
                ("box_temp", float),
                ("head_temp", float),
                ("logic", int),
                ("end", int),
            ],
        )

        for key, index in self._block_dict.items():
            if "temp" in key:
                # compute temperature
                t_value = data[:, index[0]] * 256 + data[:, index[1]]

                # something to do with the bits where you have to subtract
                t_value[np.where(t_value > 32768)] -= 65536
                value = (t_value - self.t_offset) / self.t_conversion_factor
            else:
                value = data[:, index]
            self.info_array[key][:] = value

        ### unwrap sequence
        self.info_array["sequence"] = self.unwrap_sequence(
            self.info_array["sequence"]
        )

        ### get data
        data_array = np.zeros(
            data.shape[0] * self.sample_rate,
            dtype=[
                ("hx", float),
                ("hy", float),
                ("hz", float),
                ("ex", float),
                ("ey", float),
            ],
        )

        ### fill the data
        for cc, comp in enumerate(["hx", "hy", "hz", "ex", "ey"]):
            channel_arr = np.zeros((data.shape[0], 8), dtype=float)
            for kk in range(self.sample_rate):
                index = self.indices[kk, cc]
                value = (data[:, index] * 256 + data[:, index + 1]) * np.array(
                    [256]
                ) + data[:, index + 2]
                value[np.where(value > self._int_max)] -= self._int_factor
                channel_arr[:, kk] = value
            data_array[comp][:] = channel_arr.flatten()

        ### clean things up
        ### I guess that the E channels are opposite phase?
        for comp in ["ex", "ey"]:
            data_array[comp] *= -1

        ### remove duplicates
        (
            self.info_array,
            data_array,
            self.duplicate_list,
        ) = self.remove_duplicates(self.info_array, data_array)
        ### get GPS stamps with index values
        self.stamps = self.match_status_with_gps_stamps(
            self.info_array["status"], self.gps_list
        )
        ### align data checking for timing gaps
        self.ts_data = self.align_data(data_array, self.stamps)

        et = datetime.datetime.now()
        read_time = (et - st).total_seconds()
        self.logger.debug(f"Reading took {read_time:.2f} seconds")

    def _get_first_gps_stamp(self, stamps):
        """
        get the first GPRMC stamp
        """
        for stamp in stamps:
            if stamp[1][0].gps_type in ["gprmc", "GPRMC"]:
                return stamp
        return None

    def _get_last_gps_stamp(self, stamps):
        """
        get the last gprmc stamp
        """
        for stamp in stamps[::-1]:
            if stamp[1][0].gps_type in ["gprmc", "GPRMC"]:
                return stamp
        return None

    def _locate_timing_gaps(self, stamps):
        """
        locate timing gaps in the data by comparing the stamp index with the
        GPS time stamp.  The number of points and seconds should be the same

        :param list stamps: list of GPS stamps [[status_index, [GPRMC, GPGGA]]]

        :returns: list of gap index values
        """
        stamp_01 = self._get_first_gps_stamp(stamps)[1][0]
        # current_gap = 0
        current_stamp = stamp_01
        gap_beginning = []
        total_gap = 0
        for ii, stamp in enumerate(stamps[1:], 1):
            stamp = stamp[1][0]
            # can only compare those with a date and time.
            if stamp.gps_type == "GPGGA":
                continue

            time_diff = (
                stamp.time_stamp - current_stamp.time_stamp
            ).total_seconds()
            index_diff = stamp.index - current_stamp.index

            time_gap = index_diff - time_diff
            if time_gap == 0:
                continue
            elif time_gap > 0:
                total_gap += time_gap
                current_stamp = stamp
                gap_beginning.append(stamp.index)
                self.logger.debug(
                    "GPS tamp at {0} is off from previous time by {1} seconds".format(
                        stamp.time_stamp.isoformat(),
                        time_gap,
                    )
                )

        self.logger.warning(f"Timing is off by {total_gap} seconds")
        return gap_beginning

    def check_timing(self, stamps):
        """
        make sure that there are the correct number of seconds in between
        the first and last GPS GPRMC stamps

        :param list stamps: list of GPS stamps [[status_index, [GPRMC, GPGGA]]]

        :returns: [ True | False ] if data is valid or not.
        :returns: gap index locations

        .. note:: currently it is assumed that if a data gap occurs the data can be
                  squeezed to remove them.  Probably a more elegant way of doing it.
        """
        gaps = None
        first_stamp = self._get_first_gps_stamp(stamps)[1][0]
        last_stamp = self._get_last_gps_stamp(stamps)[1][0]

        time_diff = last_stamp.time_stamp - first_stamp.time_stamp
        index_diff = last_stamp.index - first_stamp.index

        difference = index_diff - time_diff.total_seconds()
        if difference != 0:
            gaps = self._locate_timing_gaps(stamps)
            return False, gaps, difference

        return True, gaps, difference

    def align_data(self, data_array, stamps):
        """
        Need to match up the first good GPS stamp with the data

        Do this by using the first GPS stamp and assuming that the time from
        the first time stamp to the start is the index value.

        put the data into a pandas data frame that is indexed by time

        :param array data_array: structure array with columns for each
                                 component [hx, hy, hz, ex, ey]
        :param list stamps: list of GPS stamps [[status_index, [GPRMC, GPGGA]]]

        :returns: pandas DataFrame with colums of components and indexed by
                  time initialized by the start time.

        .. note:: Data gaps are squeezed cause not sure what a gap actually means.
        """
        ### check timing first to make sure there is no drift
        timing_valid, self.gaps, time_difference = self.check_timing(stamps)

        ### need to trim off the excess number of points that are present because of
        ### data gaps.  This will be the time difference times the sample rate
        if time_difference > 0:
            remove_points = int(time_difference * self.sample_rate)
            data_array = data_array[0:-remove_points]
            self.logger.info(
                f"Trimmed {remove_points} points off the end of the time "
                "series because of timing gaps"
            )

        ### first GPS stamp within the data is at a given index that is
        ### assumed to be the number of seconds from the start of the run.
        ### therefore make the start time the first GPS stamp time minus
        ### the index value for that stamp.
        ### need to be sure that the first GPS stamp has a date, need GPRMC
        first_stamp = self._get_first_gps_stamp(stamps)
        first_index = first_stamp[0]
        start_time = first_stamp[1][0].time_stamp - datetime.timedelta(
            seconds=int(first_index)
        )

        dt_index = self.make_dt_index(
            start_time.isoformat(),
            self.sample_rate,
            n_samples=data_array.shape[0],
        )

        return pd.DataFrame(data_array, index=dt_index)

    def make_dt_index(
        self, start_time, sample_rate, stop_time=None, n_samples=None
    ):
        """
        make time index array

        .. note:: date-time format should be YYYY-M-DDThh:mm:ss.ms UTC

        :param start_time: start time
        :type start_time: string

        :param end_time: end time
        :type end_time: string

        :param sample_rate: sample_rate in samples/second
        :type sample_rate: float
        """

        # set the index to be UTC time
        dt_freq = "{0:.0f}N".format(1.0 / (sample_rate) * 1e9)
        if stop_time is not None:
            dt_index = pd.date_range(
                start=start_time,
                end=stop_time,
                freq=dt_freq,
                closed="left",
                tz="UTC",
            )
        elif n_samples is not None:
            dt_index = pd.date_range(
                start=start_time, periods=n_samples, freq=dt_freq, tz="UTC"
            )
        else:
            raise ValueError("Need to input either stop_time or n_samples")

        return dt_index


# =============================================================================
# convenience read
# =============================================================================
def read_nims(fn):
    """

    :param fn: DESCRIPTION
    :type fn: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """

    nims_obj = NIMS(fn)
    nims_obj.read_nims()

    return nims_obj.to_runts()



# -*- coding: utf-8 -*-
"""
Created on Tue May 11 10:35:30 2021

:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""
# =============================================================================
# Imports
# =============================================================================
import unittest
from pathlib import Path
import numpy as np

from mth5 import mth5
from mt_metadata.timeseries.filters import PoleZeroFilter, CoefficientFilter

fn_path = Path(__file__).parent
# =============================================================================
mth5.helpers.close_open_files()


class TestFilters(unittest.TestCase):
    """
    Test filters to make sure get out what is put in
    """

    @classmethod
    def setUpClass(self):
        self.fn = fn_path.joinpath("filter_test.h5")
        self.m_obj = mth5.MTH5(file_version="0.1.0")
        self.m_obj.open_mth5(self.fn, "w")
        self.filter_group = self.m_obj.filters_group

        self.zpk = PoleZeroFilter()
        self.zpk.units_in = "counts"
        self.zpk.units_out = "mv"
        self.zpk.name = "zpk_test"
        self.zpk.poles = np.array([1 + 2j, 0, 1 - 2j])
        self.zpk.zeros = np.array([10 - 1j, 10 + 1j])

        self.coefficient = CoefficientFilter()
        self.coefficient.units_in = "volts"
        self.coefficient.units_out = "millivolts per meter"
        self.coefficient.name = "coefficient_test"
        self.coefficient.gain = 10.0

        self.zpk_group = self.filter_group.add_filter(self.zpk)
        self.coefficient_group = self.filter_group.add_filter(self.coefficient)

    def test_zpk_in(self):

        self.assertIn("zpk_test", self.filter_group.zpk_group.groups_list)

    def test_zpk_name(self):
        self.assertEqual(self.zpk_group.attrs["name"], self.zpk.name)

    def test_zpk_units_in(self):
        self.assertEqual(self.zpk_group.attrs["units_in"], self.zpk.units_in)

    def test_zpk_units_out(self):
        self.assertEqual(self.zpk_group.attrs["units_out"], self.zpk.units_out)

    def test_zpk_poles(self):
        self.assertTrue(
            np.allclose(self.zpk_group["poles"][()], self.zpk.poles)
        )

    def test_zpk_zeros(self):
        self.assertTrue(np.allclose(self.zpk_group["zeros"], self.zpk.zeros))

    def test_zpk_out(self):
        new_zpk = self.filter_group.to_filter_object(self.zpk.name)

        self.assertTrue(new_zpk == self.zpk)

    def test_coefficient_in(self):

        self.assertIn(
            "coefficient_test", self.filter_group.coefficient_group.groups_list
        )

    def test_coefficient_name(self):
        self.assertEqual(
            self.coefficient_group.attrs["name"], self.coefficient.name
        )

    def test_coefficient_units_in(self):
        self.assertEqual(
            self.coefficient_group.attrs["units_in"], self.coefficient.units_in
        )

    def test_coefficient_units_out(self):
        self.assertEqual(
            self.coefficient_group.attrs["units_out"],
            self.coefficient.units_out,
        )

    def test_coefficient_out(self):
        new_coefficient = self.filter_group.to_filter_object(
            self.coefficient.name
        )

        self.assertTrue(new_coefficient == self.coefficient)

    @classmethod
    def tearDownClass(self):
        self.m_obj.close_mth5()
        self.fn.unlink()



# -*- coding: utf-8 -*-
"""
Phoenix file collection

Created on Thu Aug  4 16:48:47 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
from collections import OrderedDict
from pathlib import Path

import pandas as pd

from mth5.utils.mth5_logger import setup_logger


# =============================================================================


class Collection:
    """
    A general collection class to keep track of files with methods to create
    runs and run ids.

    """

    def __init__(self, file_path=None, **kwargs):

        self.logger = setup_logger(f"{__name__}.{self.__class__.__name__}")
        self.file_path = file_path
        self.file_ext = "*"

        self._columns = [
            "survey",
            "station",
            "run",
            "start",
            "end",
            "channel_id",
            "component",
            "fn",
            "sample_rate",
            "file_size",
            "n_samples",
            "sequence_number",
            "instrument_id",
            "calibration_fn",
        ]

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        lines = [
            f"Collection for file type {self.file_ext} in {self._file_path}"
        ]

        return "\n".join(lines)

    def __repr__(self):
        return f"Collection({self.file_path})"

    @property
    def file_path(self):
        """
        Path object to file directory
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        """
        :param file_path: path to files
        :type file_path: string or Path object

        sets file_path as a Path object
        """

        if file_path is None:
            self._file_path = None
        if not isinstance(file_path, Path):
            file_path = Path(file_path)

        self._file_path = file_path

        if not self._file_path.exists():
            raise IOError()

    def get_files(self, extension):
        """
        Get files with given extension. Uses Pathlib.Path.rglob, so it finds
        all files within the `file_path` by searching all sub-directories.

        :param extension: file extension(s)
        :type extension: string or list
        :return: list of files in the `file_path` with the given extensions
        :rtype: list of Path objects

        """

        if isinstance(extension, (list, tuple)):
            fn_list = []
            for ext in extension:
                fn_list += list(self.file_path.rglob(f"*.{ext}"))

        else:
            fn_list = list(self.file_path.rglob(f"*.{extension}"))

        return fn_list

    def to_dataframe(self):
        """
        Get a data frame of the file summary with column names:

            - **survey**: survey id
            - **station**: station id
            - **run**: run id
            - **start**: start time UTC
            - **end**: end time UTC
            - **channel_id**: channel id or list of channel id's in file
            - **component**: channel component or list of components in file
            - **fn**: path to file
            - **sample_rate**: sample rate in samples per second
            - **file_size**: file size in bytes
            - **n_samples**: number of samples in file
            - **sequence_number**: sequence number of the file
            - **instrument_id**: instrument id
            - **calibration_fn**: calibration file

        :return: summary table of file names,
        :rtype: TYPE

        """
        pass

    def assign_run_names(self):
        """

        :return: DESCRIPTION
        :rtype: TYPE

        """
        pass

    def _set_df_dtypes(self, df):
        """
        Set some of the columns in the dataframe to desired types

            - **start**: pandas.datetime
            - **end**: pandas.datetime
            - **instrument_id**: string
            - **calibration_fn**: string

        :param df: summary table
        :type df: :class:`pandas.DataFrame`
        :return: summary table with proper types
        :rtype: :class:`pandas.DataFrame`

        """

        df.start = pd.to_datetime(df.start, errors="coerce")
        df.end = pd.to_datetime(df.end, errors="coerce")
        df.instrument_id = df.instrument_id.astype(str)
        df.calibration_fn = df.calibration_fn.astype(str)

        return df

    def _sort_df(self, df, zeros):
        """
        sort to a given dataframe by start date and then by run name. The
        index is reset.

        :param df: summary table
        :type df: :class:`pandas.DataFrame`
        :param zeros: number of zeros in run id
        :type zeros: integer
        :return: summary table sorted by start time and run id
        :rtype: :class:`pandas.DataFrame`

        """

        df.sort_values(by=["start"], inplace=True)

        # assign run names
        df = self.assign_run_names(df, zeros=zeros)

        df.sort_values(by=["run", "start"], inplace=True)
        df.reset_index(inplace=True, drop=True)

        return df

    def get_runs(
        self,
        sample_rates,
        run_name_zeros=4,
        calibration_path=None,
    ):
        """
        Get a list of runs contained within the given folder.  First the
        dataframe will be developed from which the runs are extracted.

        For continous data all you need is the first file in the sequence. The
        reader will read in the entire sequence.

        For segmented data it will only read in the given segment, which is
        slightly different from the original reader.

        :param sample_rates: list of sample rates to read, defaults to [150, 24000]
        :param run_name_zeros: Number of zeros in the run name, defaults to 4
        :type run_name_zeros: integer, optional
        :return: List of run dataframes with only the first block of files
        :rtype: :class:`collections.OrderedDict`

        :Example:

            >>> from mth5.io.phoenix import PhoenixCollection
            >>> phx_collection = PhoenixCollection(r"/path/to/station")
            >>> run_dict = phx_collection.get_runs(sample_rates=[150, 24000])

        """

        df = self.to_dataframe(
            sample_rates=sample_rates,
            run_name_zeros=run_name_zeros,
            calibration_path=calibration_path,
        )

        run_dict = OrderedDict()

        for station in sorted(df.station.unique()):
            run_dict[station] = OrderedDict()

            for run_id in sorted(
                df[df.station == station].run.unique(),
                key=lambda x: x[-run_name_zeros:],
            ):
                run_df = df[(df.station == station) & (df.run == run_id)]
                run_dict[station][run_id] = run_df

        return run_dict
